class Light(object):
    '''Light facade.

    Light sensor measures the ambient light level(illumination) in lx.
    Common uses include controlling screen brightness.

    With method `enable` you can turn on the sensor and
    `disable` method stops the sensor.

    Use property `illumination` to get current illumination in lx.

    .. versionadded:: 1.2.5

    Supported Platforms:: Android
    '''

    @property
    def illumination(self):
        '''Current illumination in lx.'''
        return self._get_illumination()

    def enable(self):
        '''Enable light sensor.'''
        self._enable()

    def disable(self):
        '''Disable light sensor.'''
        self._disable()

    # private

    def _get_illumination(self, **kwargs):
        raise NotImplementedError()

    def _enable(self, **kwargs):
        raise NotImplementedError()

    def _disable(self, **kwargs):
        raise NotImplementedError()



'''
Email
=====

The :class:`Email` provides access to public methods to use email of your
device.

.. note::
    On Android `INTERNET` permission is needed.

Simple Examples
---------------

To send an e-mail::

    >>> from plyer import email
    >>> recipient = 'abc@gmail.com'
    >>> subject = 'Hi'
    >>> text = 'This is an example.'
    >>> create_chooser = False
    >>> email.send(recipient=recipient, subject=subject, text=text,
                   create_chooser=create_chooser)

    >>> # opens email interface where user can change the content.

Supported Platforms
-------------------
Android, iOS, Windows, OS X, Linux

'''


class Email(object):
    # pylint: disable=too-few-public-methods
    '''
    Email facade.
    '''

    def send(self, recipient=None, subject=None, text=None,
             create_chooser=None):
        '''
        Open an email client message send window, prepopulated with the
        given arguments.

        :param recipient: Recipient of the message (str)
        :param subject: Subject of the message (str)
        :param text: Main body of the message (str)
        :param create_chooser: Whether to display a program chooser to
                               handle the message (bool)

        .. note:: create_chooser is only supported on Android
        '''
        self._send(recipient=recipient, subject=subject, text=text,
                   create_chooser=create_chooser)

    # private

    def _send(self, **kwargs):
        raise NotImplementedError()



'''
iOS Storage Path
--------------------
'''

from plyer.facades import StoragePath
from pyobjus import autoclass
import os

NSFileManager = autoclass('NSFileManager')

# Directory constants (NSSearchPathDirectory enumeration)
NSApplicationDirectory = 1
NSDocumentDirectory = 9
NSDownloadsDirectory = 15
NSMoviesDirectory = 17
NSMusicDirectory = 18
NSPicturesDirectory = 19


class iOSStoragePath(StoragePath):

    def __init__(self):
        self.defaultManager = NSFileManager.defaultManager()

    def _get_home_dir(self):
        return os.path.expanduser('~/')

    def _get_external_storage_dir(self):
        return 'This feature is not implemented for this platform.'

    def _get_root_dir(self):
        return 'This feature is not implemented for this platform.'

    def _get_documents_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSDocumentDirectory, 1).firstObject().absoluteString.UTF8String()

    def _get_downloads_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSDownloadsDirectory, 1).firstObject().absoluteString.UTF8String()

    def _get_videos_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSMoviesDirectory, 1).firstObject().absoluteString.UTF8String()

    def _get_music_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSMusicDirectory, 1).firstObject().absoluteString.UTF8String()

    def _get_pictures_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSPicturesDirectory, 1).firstObject().absoluteString.UTF8String()

    def _get_application_dir(self):
        return self.defaultManager.URLsForDirectory_inDomains_(
            NSApplicationDirectory, 1).firstObject().absoluteString.\
            UTF8String()


def instance():
    return iOSStoragePath()



'''
Module of Windows API for plyer.notification.
'''

from threading import Thread as thread

from plyer.facades import Notification
from plyer.platforms.win.libs.balloontip import balloon_tip


class WindowsNotification(Notification):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of Windows notification/balloon API.
    '''

    def _notify(self, **kwargs):
        thread(target=balloon_tip, kwargs=kwargs).start()


def instance():
    '''
    Instance for facade proxy.
    '''
    return WindowsNotification()



'''
Call
====

The :class:`Call` provides access to calling feature of your device.

.. note::
    - On Android your app needs the `CALL_PHONE` or `CALL_PRIVILEGED`
    permission in order to make calls.

    - Dialing call feature in not supported yet in iOS devices.

Simple Examples
---------------

To make call::

    >>> from plyer import call
    >>> tel = 9999222299
    >>> call.makecall(tel=tel)

To dial call::

    >>> call.dialcall()

Supported Platforms
-------------------
Android, iOS

'''


class Call(object):
    '''
    Call facade.
    '''

    def makecall(self, tel):
        '''
        Make calls using your device.

        :param tel: The reciever
        :type tel: number
        '''
        self._makecall(tel=tel)

    def dialcall(self):
        '''
        Opens dialing interface.
        '''
        self._dialcall()

    # private

    def _makecall(self, **kwargs):
        raise NotImplementedError()

    def _dialcall(self, **kwargs):
        raise NotImplementedError()



'''
Module of MacOS API for plyer.uniqueid.
'''

from os import environ
from subprocess import Popen, PIPE
from plyer.facades import UniqueID
from plyer.utils import whereis_exe


class OSXUniqueID(UniqueID):
    '''
    Implementation of MacOS uniqueid API.
    '''

    def _get_uid(self):
        old_lang = environ.get('LANG')
        environ['LANG'] = 'C'

        ioreg_process = Popen(["ioreg", "-l"], stdout=PIPE)
        grep_process = Popen(
            ["grep", "IOPlatformSerialNumber"],
            stdin=ioreg_process.stdout, stdout=PIPE
        )
        ioreg_process.stdout.close()
        output = grep_process.communicate()[0]

        if old_lang is None:
            environ.pop('LANG')
        else:
            environ['LANG'] = old_lang

        result = None
        if output:
            result = output.split()[3][1:-1]
        return result


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('ioreg'):
        return OSXUniqueID()
    sys.stderr.write("ioreg not found.")
    return UniqueID()



'''
Orientation
==========

The :class:`Orientation` provides access to public methods to set orientation
of your device.

.. note::
    These settings are generally guidelines, the operating
    system may choose to ignore them, or they may be overridden by
    other system components.

.. versionadded:: 1.2.4

Simple Examples
---------------

To set landscape::

    >>> from plyer import orientation
    >>> orientation.set_landscape()

To set portrait::

    >>> orientation.set_portrait()

To set sensor::

    >>> orientation.set_sensor()

Supported Platforms
-------------------
Android

'''


class Orientation(object):
    '''
    Orientation facade.
    '''

    def set_landscape(self, reverse=False):
        '''
        Rotate the app to a landscape orientation.

        :param reverse: If True, uses the opposite of the natural
                        orientation.
        '''
        self._set_landscape(reverse=reverse)

    def set_portrait(self, reverse=False):
        '''
        Rotate the app to a portrait orientation.

        :param reverse: If True, uses the opposite of the natural
                        orientation.
        '''
        self._set_portrait(reverse=reverse)

    def set_sensor(self, mode='any'):
        '''
        Rotate freely following sensor information from the device.

        :param mode: The rotation mode, should be one of 'any' (rotate
                     to any orientation), 'landscape' (choose nearest
                     landscape mode) or 'portrait' (choose nearest
                     portrait mode). Defaults to 'any'.
        '''
        self._set_sensor(mode=mode)

    # private

    def _set_landscape(self, **kwargs):
        raise NotImplementedError()

    def _set_portrait(self, **kwargs):
        raise NotImplementedError()

    def _set_sensor(self, **kwargs):
        raise NotImplementedError()



'''
Utils
=====

'''
# pylint: disable=useless-object-inheritance

__all__ = ('platform', 'reify', 'deprecated')

from os import environ
from os import path
from sys import platform as _sys_platform


class Platform(object):
    '''
    Refactored to class to allow module function to be replaced
    with module variable.
    '''

    def __init__(self):
        self._platform_ios = None
        self._platform_android = None

    def __eq__(self, other):
        return other == self._get_platform()

    def __ne__(self, other):
        return other != self._get_platform()

    def __str__(self):
        return self._get_platform()

    def __repr__(self):
        return 'platform name: \'{platform}\' from: \n{instance}'.format(
            platform=self._get_platform(),
            instance=super(Platform, self).__repr__()
        )

    def __hash__(self):
        return self._get_platform().__hash__()

    def _get_platform(self):

        if self._platform_android is None:
            # ANDROID_ARGUMENT and ANDROID_PRIVATE are 2 environment variables
            # from python-for-android project
            self._platform_android = 'ANDROID_ARGUMENT' in environ

        if self._platform_ios is None:
            self._platform_ios = (environ.get('KIVY_BUILD', '') == 'ios')

        # On android, _sys_platform return 'linux2', so prefer to check the
        # import of Android module than trying to rely on _sys_platform.

        # pylint: disable=no-else-return
        if self._platform_android is True:
            return 'android'
        elif self._platform_ios is True:
            return 'ios'
        elif _sys_platform in ('win32', 'cygwin'):
            return 'win'
        elif _sys_platform == 'darwin':
            return 'macosx'
        elif _sys_platform[:5] == 'linux':
            return 'linux'
        return 'unknown'


platform = Platform()  # pylint: disable=invalid-name


class Proxy(object):
    '''
    Based on http://code.activestate.com/recipes/496741-object-proxying
    version by Tomer Filiba, PSF license.
    '''

    __slots__ = ['_obj', '_name', '_facade']

    def __init__(self, name, facade):
        object.__init__(self)
        object.__setattr__(self, '_obj', None)
        object.__setattr__(self, '_name', name)
        object.__setattr__(self, '_facade', facade)

    def _ensure_obj(self):
        obj = object.__getattribute__(self, '_obj')
        if obj:
            return obj
        # do the import
        try:
            name = object.__getattribute__(self, '_name')
            module = 'plyer.platforms.{}.{}'.format(
                platform, name)
            mod = __import__(module, fromlist='.')
            obj = mod.instance()
        except:  # pylint: disable=bare-except
            import traceback
            traceback.print_exc()
            facade = object.__getattribute__(self, '_facade')
            obj = facade()

        object.__setattr__(self, '_obj', obj)
        return obj

    def __getattribute__(self, name):
        result = None

        if name == '__doc__':
            return result

        # run _ensure_obj func, result in _obj
        object.__getattribute__(self, '_ensure_obj')()

        # return either Proxy instance or platform-dependent implementation
        result = getattr(object.__getattribute__(self, '_obj'), name)
        return result

    def __delattr__(self, name):
        object.__getattribute__(self, '_ensure_obj')()
        delattr(object.__getattribute__(self, '_obj'), name)

    def __setattr__(self, name, value):
        object.__getattribute__(self, '_ensure_obj')()
        setattr(object.__getattribute__(self, '_obj'), name, value)

    def __bool__(self):
        object.__getattribute__(self, '_ensure_obj')()
        return bool(object.__getattribute__(self, '_obj'))

    def __str__(self):
        object.__getattribute__(self, '_ensure_obj')()
        return str(object.__getattribute__(self, '_obj'))

    def __repr__(self):
        object.__getattribute__(self, '_ensure_obj')()
        return repr(object.__getattribute__(self, '_obj'))


def whereis_exe(program):
    ''' Tries to find the program on the system path.
        Returns the path if it is found or None if it's not found.
    '''
    path_split = ';' if platform == 'win' else ':'
    for pth in environ.get('PATH', '').split(path_split):
        folder = path.isdir(path.join(pth, program))
        available = path.exists(path.join(pth, program))
        if available and not folder:
            return path.join(pth, program)
    return None


class reify(object):
    # pylint: disable=too-few-public-methods,invalid-name
    '''
    Put the result of a method which uses this (non-data) descriptor decorator
    in the instance dict after the first call, effectively replacing the
    decorator with an instance variable.

    It acts like @property, except that the function is only ever called once;
    after that, the value is cached as a regular attribute. This gives you lazy
    attribute creation on objects that are meant to be immutable.

    Taken from the `Pyramid project <https://pypi.python.org/pypi/pyramid/>`_.

    To use this as a decorator::

         @reify
         def lazy(self):
              ...
              return hard_to_compute_int
         first_time = self.lazy   # lazy is reify obj, reify.__get__() runs
         second_time = self.lazy  # lazy is hard_to_compute_int
    '''

    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        retval = self.func(inst)
        setattr(inst, self.func.__name__, retval)
        return retval


def deprecated(obj):
    '''
    This is a decorator which can be used to mark functions and classes as
    deprecated. It will result in a warning being emitted when a deprecated
    function is called or a new instance of a class created.

    In case of classes, the warning is emitted before the __new__ method
    of the decorated class is called, therefore a way before the __init__
    method itself.
    '''

    import warnings
    from inspect import stack
    from functools import wraps
    from types import FunctionType, MethodType

    new_obj = None

    # wrap a function into a function emitting a deprecated warning
    if isinstance(obj, FunctionType):

        @wraps(obj)
        def new_func(*args, **kwargs):
            # get the previous stack frame and extract file, line and caller
            # stack() -> caller()
            call_file, call_line, caller = stack()[1][1:4]

            # assemble warning
            warning = (
                'Call to deprecated function {} in {} line {}. '
                'Called from {} line {}'
                ' by {}().\n'.format(
                    obj.__name__,
                    obj.__code__.co_filename,
                    obj.__code__.co_firstlineno + 1,
                    call_file, call_line, caller
                )
            )

            warnings.warn('[{}] {}'.format('WARNING', warning))

            # if there is a docstring present, emit docstring too
            if obj.__doc__:
                warnings.warn(obj.__doc__)

            # return function wrapper
            return obj(*args, **kwargs)
        new_obj = new_func

    # wrap a class into a class emitting a deprecated warning
    # obj is class, type(obj) is metaclass, metaclasses inherit from type
    elif isinstance(type(obj), type):
        # we have an access to the metaclass instance (class) and need to print
        # the warning when a class instance (object) is created with __new__
        # i.e. when calling Class()

        def obj_new(cls, child, *args, **kwargs):
            '''
            Custom metaclass instance's __new__ method with deprecated warning.
            Calls the original __new__ method afterwards.
            '''
            # pylint: disable=unused-argument

            # get the previous stack frame and extract file, line and caller
            # stack() -> caller()
            call_file, call_line, caller = stack()[1][1:4]
            loc_file = obj.__module__

            warnings.warn(
                '[{}] Creating an instance of a deprecated class {} in {}.'
                ' Called from {} line {} by {}().\n'.format(
                    'WARNING', obj.__name__, loc_file,
                    call_file, call_line, caller
                )
            )

            # if there is a docstring present, emit docstring too
            if obj.__doc__:
                warnings.warn(obj.__doc__)

            # make sure nothing silly gets into the function
            assert obj is cls

            # we are creating a __new__ for a class that inherits from
            # a deprecated class, therefore in this particular case
            # MRO is (child, cls, object) > (cls, object)
            if len(child.__mro__) > len(cls.__mro__):
                assert cls is child.__mro__[1], (cls.__mro__, child.__mro__)

            # we are creating __new__ directly for the deprecated class
            # therefore MRO is the same for parent and child class
            elif len(child.__mro__) == len(cls.__mro__):
                assert cls is child

            # return the class back with the extended __new__ method
            return obj.__old_new__(child)

        # back up the old __new__ method and create an extended
        # __new__ method that emits deprecated warnings
        obj.__old_new__ = obj.__new__
        obj.__new__ = MethodType(obj_new, obj)
        new_obj = obj

    # return a function wrapper or an extended class
    return new_obj



class Proximity(object):
    '''Proximity facade.

    The proximity sensor is commonly used to determine distance whether
    phone is close to your head. Commonly is used when you have a call
    and you stick your phone with your head. Then screen of phone turns off.

    Use method `enable` to turn on proximity sensor and method `disable` for
    turn off.

    To check if some object (or your head) is near sensor check values from
    property `proximity`. It returns `True` when object is close.

    .. versionadded:: 1.2.5

    Supported Platforms::Android
    '''

    @property
    def proximity(self):
        '''Return True or False depending if there is an object or not.

        :return: True if there is an object. Otherwise False.
        '''
        return self._get_proximity()

    def _enable(self, **kwargs):
        raise NotImplementedError()

    def enable(self):
        '''Enable the proximity sensor.
        '''
        self._enable()

    def _disable(self, **kwargs):
        raise NotImplementedError()

    def disable(self):
        '''Disable the proximity sensor.
        '''
        self._disable()

    def _get_proximity(self):
        raise NotImplementedError()



'''
Facades
=======

Interface of all the features available.

'''

__all__ = ('Accelerometer', 'Audio', 'Barometer', 'Battery', 'Call', 'Camera',
           'Compass', 'Email', 'FileChooser', 'GPS', 'Gravity', 'Gyroscope',
           'IrBlaster', 'Light', 'Orientation', 'Notification', 'Proximity',
           'Sms', 'TTS', 'UniqueID', 'Vibrator', 'Wifi', 'Flash', 'CPU',
           'Temperature', 'Humidity', 'SpatialOrientation', 'Brightness',
           'Processors', 'StoragePath', 'keystore', 'Bluetooth', 'Screenshot',
           'STT')

from plyer.facades.accelerometer import Accelerometer
from plyer.facades.audio import Audio
from plyer.facades.barometer import Barometer
from plyer.facades.battery import Battery
from plyer.facades.call import Call
from plyer.facades.camera import Camera
from plyer.facades.compass import Compass
from plyer.facades.email import Email
from plyer.facades.filechooser import FileChooser
from plyer.facades.flash import Flash
from plyer.facades.gps import GPS
from plyer.facades.gravity import Gravity
from plyer.facades.gyroscope import Gyroscope
from plyer.facades.irblaster import IrBlaster
from plyer.facades.light import Light
from plyer.facades.proximity import Proximity
from plyer.facades.orientation import Orientation
from plyer.facades.notification import Notification
from plyer.facades.sms import Sms
from plyer.facades.stt import STT
from plyer.facades.tts import TTS
from plyer.facades.uniqueid import UniqueID
from plyer.facades.vibrator import Vibrator
from plyer.facades.wifi import Wifi
from plyer.facades.temperature import Temperature
from plyer.facades.humidity import Humidity
from plyer.facades.spatialorientation import SpatialOrientation
from plyer.facades.brightness import Brightness
from plyer.facades.keystore import Keystore
from plyer.facades.storagepath import StoragePath
from plyer.facades.bluetooth import Bluetooth
from plyer.facades.processors import Processors
from plyer.facades.cpu import CPU
from plyer.facades.screenshot import Screenshot



'''
Module of MacOS API for plyer.cpu.
'''

from subprocess import Popen, PIPE
from plyer.facades import CPU
from plyer.utils import whereis_exe


class OSXCPU(CPU):
    '''
    Implementation of MacOS CPU API.
    '''

    @staticmethod
    def _sockets():
        return

    def _physical(self):
        # cores
        physical = None

        _physical = Popen(
            ['sysctl', '-n', 'hw.physicalcpu_max'],
            stdout=PIPE
        )
        output = _physical.communicate()[0].decode('utf-8').strip()
        if output:
            physical = int(output)
        return physical

    def _logical(self):
        # cores * threads
        logical = None

        _logical = Popen(
            ['sysctl', '-n', 'hw.logicalcpu_max'],
            stdout=PIPE
        )
        output = _logical.communicate()[0].decode('utf-8').strip()
        if output:
            logical = int(output)
        return logical

    @staticmethod
    def _cache():
        return

    @staticmethod
    def _numa():
        return


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('sysctl'):
        return OSXCPU()
    sys.stderr.write('sysctl not found.')
    return CPU()



'''
Module of Linux API for plyer.cpu.
'''

from os.path import join
from os import environ, listdir
from subprocess import Popen, PIPE
from plyer.facades import CPU
from plyer.utils import whereis_exe


class LinuxCPU(CPU):
    '''
    Implementation of Linux CPU API.
    '''

    def _sockets(self):
        # physical CPU sockets (or slots) on motherboard
        sockets = []  # list of CPU ids from kernel

        # open Linux kernel data file for CPU
        with open('/proc/cpuinfo', 'rb') as fle:
            lines = fle.readlines()

        # go through the lines and obtain physical CPU ids
        for line in lines:
            line = line.decode('utf-8')
            if 'physical id' not in line:
                continue
            cpuid = line.split(':')[1].strip()
            sockets.append(cpuid)

        # total sockets is the length of unique CPU ids from kernel
        sockets = len(set(sockets))
        return sockets

    def _physical(self):
        # cores
        physical = []  # list of CPU ids from kernel

        # open Linux kernel data file for CPU
        with open('/proc/cpuinfo', 'rb') as fle:
            lines = fle.readlines()

        # go through the lines and obtain CPU core ids
        for line in lines:
            line = line.decode('utf-8')
            if 'core id' not in line:
                continue
            cpuid = line.split(':')[1].strip()
            physical.append(cpuid)

        # total cores (socket * core per socket)
        # is the length of unique CPU core ids from kernel
        physical = len(set(physical))
        return physical

    def _logical(self):
        # cores * threads
        logical = None
        old_lang = environ.get('LANG', '')
        environ['LANG'] = 'C'

        _logical = Popen(['nproc', '--all'], stdout=PIPE)
        output = _logical.communicate()[0].decode('utf-8').strip()
        if output:
            logical = int(output)

        environ['LANG'] = old_lang
        return logical

    def _cache(self):
        values = {key: 0 for key in ('L1', 'L2', 'L3')}
        cpu_path = join('/sys', 'devices', 'system', 'cpu')

        # get present cores from kernel device
        with open(join(cpu_path, 'present')) as fle:
            present = fle.read().decode('utf-8')
        present = present.strip().split('-')

        if len(present) == 2:
            present = range(int(present[1]) + 1)
        else:
            present = [present[0]]

        cores = ['cpu{}'.format(i) for i in present]
        for core in cores:
            indicies = [
                # get 'indexN' files from 'cache' folder assuming
                # the filename is in range index0 to index99
                # in case a wild 'index_whatevercontent' file appears
                fle for fle in listdir(join(cpu_path, core, 'cache'))
                if fle.startswith('index') and len(fle) <= len('index') + 2
            ]

            for index in indicies:
                index_type = join(cpu_path, core, 'cache', index, 'level')
                with open(index_type, 'rb') as fle:
                    cache_level = fle.read().decode('utf-8').strip()
                values['L{}'.format(cache_level)] += 1
        return values

    @staticmethod
    def _numa():
        return


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('nproc'):
        return LinuxCPU()
    sys.stderr.write("nproc not found.")
    return CPU()



'''
GPS
====

.. versionadded:: 1.1

.. note::
    On Android `INTERNET`, `ACCESS_FINE_LOCATION`, `ACCESS_COARSE_LOCATION`
    permissions are needed.


.. note::
    On iOS `NSLocationWhenInUseUsageDescription` key is required for app to
    display geolocation usage permission prompt. Key can be added in Xcode
    target `info` section or in ``Resources/<YourApp>-info.plist``.
    App background mode (`on_pause`) also must be supported.

You need to set a `on_location` callback with the :meth:`configure` method.
This callback will receive a couple of keywords / values, that might be
different depending of their availability on the targeted platform.
Lat and lon are always available.

- lat: latitude of the last location, in degrees
- lon: longitude of the last location, in degrees
- speed: speed of the user, in meters/second over ground
- bearing: bearing in degrees
- altitude: altitude in meters above the sea level

Here is an example of the usage of gps::

    from plyer import gps

    def print_locations(**kwargs):
        print 'lat: {lat}, lon: {lon}'.format(**kwargs)

    gps.configure(on_location=print_locations)
    gps.start()
    # later
    gps.stop()

Supported Platforms
-------------------
Android, iOS

'''


class GPS(object):
    '''
    GPS facade.
    '''

    def configure(self, on_location, on_status=None):
        '''
        Configure the GPS object. This method should be called before
        :meth:`start`.

        :param on_location: Function to call when receiving a new location
        :param on_status: Function to call when a status message is received
        :type on_location: callable, multiples keys/value will be passed.
        :type on_status: callable, args are "message-type", "status"

        .. warning::

            The `on_location` and `on_status` callables might be called from
            another thread than the thread used for creating the GPS object.
        '''
        self.on_location = on_location
        self.on_status = on_status
        self._configure()

    def start(self, minTime=1000, minDistance=1):
        '''
        Start the GPS location updates.
        Expects 2 parameters:
            minTime: milliseconds.  (float)
            minDistance: meters. (float)
        '''
        self._start(minTime=minTime, minDistance=minDistance)

    def stop(self):
        '''
        Stop the GPS location updates.
        '''
        self._stop()

    # private

    def _configure(self):
        raise NotImplementedError()

    def _start(self, **kwargs):
        raise NotImplementedError()

    def _stop(self):
        raise NotImplementedError()



'''
Module of iOS API for plyer.battery.
'''

from pyobjus import autoclass  # pylint: disable=import-error
from pyobjus.dylib_manager import load_framework  # pylint:disable=import-error
from plyer.facades import Battery

load_framework('/System/Library/Frameworks/UIKit.framework')
UIDevice = autoclass('UIDevice')


class IOSBattery(Battery):
    '''
    Implementation of iOS battery API.
    '''

    def __init__(self):
        super(IOSBattery, self).__init__()
        self.device = UIDevice.currentDevice()

    def _get_state(self):
        status = {"isCharging": None, "percentage": None}

        if not self.device.batteryMonitoringEnabled:
            self.device.setBatteryMonitoringEnabled_(True)

        if self.device.batteryState == 0:
            is_charging = None
        elif self.device.batteryState == 2:
            is_charging = True
        else:
            is_charging = False

        percentage = self.device.batteryLevel * 100.

        status['isCharging'] = is_charging
        status['percentage'] = percentage

        return status


def instance():
    '''
    Instance for facade proxy.
    '''
    return IOSBattery()



class Gravity(object):
    '''Gravity facade.

    .. versionadded:: 1.2.5

    Supported Platforms:: Android

    '''

    @property
    def gravity(self):
        '''Property that returns values of the current gravity force
        as a (x, y, z) tuple. Returns (None, None, None)
        if no data is currently available.
        '''
        return self._get_gravity()

    def enable(self):
        '''Activate the gravity sensor. Throws an error if the
        hardware is not available or not implemented on.
        '''
        self._enable()

    def disable(self):
        '''Disable the gravity sensor.
        '''
        self._disable()

    # private

    def _enable(self):
        raise NotImplementedError()

    def _disable(self):
        raise NotImplementedError()

    def _get_gravity(self):
        raise NotImplementedError()



'''
Module of Windows API for plyer.uniqueid.
'''

try:
    import _winreg as regedit  # pylint: disable=import-error
except ImportError:
    try:
        import winreg as regedit  # pylint: disable=import-error
    except ImportError:
        raise NotImplementedError()

from plyer.facades import UniqueID


class WinUniqueID(UniqueID):
    '''
    Implementation of Windows battery API.
    '''

    def _get_uid(self):
        # Win XP+, REG QUERY KEY /V VALUE, case-insensitive
        handle = regedit.OpenKey(
            regedit.HKEY_LOCAL_MACHINE,
            r"SOFTWARE\\Microsoft\\Cryptography", 0,
            regedit.KEY_READ | regedit.KEY_WOW64_64KEY
        )
        value, _ = regedit.QueryValueEx(handle, "MachineGuid")
        return value


def instance():
    '''
    Instance for facade proxy.
    '''
    return WinUniqueID()



# coding=utf-8


class Temperature(object):
    '''Temperature facade.

    Temperature sensor is used to measure the ambient room temperature in
    degrees Celsius (°C)
    With method `enable` you can turn on temperature sensor and 'disable'
    method stops the sensor.
    Use property `temperature` to get ambient air temperature in degree C.

    .. versionadded:: 1.2.5

    Supported Platforms:: Android

    '''

    @property
    def temperature(self):
        '''Current air temperature in degree C.'''
        return self._get_temperature()

    def enable(self):
        '''Enable temperature sensor.'''
        self._enable()

    def disable(self):
        '''Disable temperature sensor.'''
        self._disable()

    # private

    def _get_temperature(self, **kwargs):
        raise NotImplementedError()

    def _enable(self, **kwargs):
        raise NotImplementedError()

    def _disable(self, **kwargs):
        raise NotImplementedError()



class Barometer(object):
    '''Barometer facade.

    Barometer sensor is used to measure the ambient air pressure in hPa.

    With method `enable` you can turn on pressure sensor and 'disable'
    method stops the sensor.

    Use property `pressure` to get current air pressure in hPa.

    .. versionadded:: 1.2.5

    Supported Platforms:: Android
    '''

    @property
    def pressure(self):
        '''Current air pressure in hPa.'''
        return self._get_pressure()

    def _get_pressure(self, **kwargs):
        raise NotImplementedError()

    def _enable(self, **kwargs):
        raise NotImplementedError()

    def enable(self):
        '''Enable barometer sensor.'''
        self._enable()

    def _disable(self, **kwargs):
        raise NotImplementedError()

    def disable(self):
        '''Disable barometer sensor.'''
        self._disable()



"""Implementation Vibrator for Android."""

from jnius import autoclass, cast
from plyer.facades import Vibrator
from plyer.platforms.android import activity
from plyer.platforms.android import SDK_INT

Context = autoclass("android.content.Context")
vibrator_service = activity.getSystemService(Context.VIBRATOR_SERVICE)
vibrator = cast("android.os.Vibrator", vibrator_service)
if SDK_INT >= 26:
    VibrationEffect = autoclass("android.os.VibrationEffect")


class AndroidVibrator(Vibrator):
    """Android Vibrator class.

    Supported features:
        * vibrate for some period of time.
        * vibrate from given pattern.
        * cancel vibration.
        * check whether Vibrator exists.
    """

    def _vibrate(self, time=None, **kwargs):
        if vibrator:
            if SDK_INT >= 26:
                vibrator.vibrate(
                    VibrationEffect.createOneShot(
                        int(1000 * time), VibrationEffect.DEFAULT_AMPLITUDE
                    )
                )
            else:
                vibrator.vibrate(int(1000 * time))

    def _pattern(self, pattern=None, repeat=None, **kwargs):
        pattern = [int(1000 * time) for time in pattern]

        if vibrator:
            if SDK_INT >= 26:
                vibrator.vibrate(
                    VibrationEffect.createWaveform(pattern, repeat)
                )
            else:
                vibrator.vibrate(pattern, repeat)

    def _exists(self, **kwargs):
        if SDK_INT >= 11:
            return vibrator.hasVibrator()
        elif vibrator_service is None:
            raise NotImplementedError()
        return True

    def _cancel(self, **kwargs):
        vibrator.cancel()


def instance():
    """Returns Vibrator with android features.

    :return: instance of class AndroidVibrator
    """
    return AndroidVibrator()



'''
Screenshot
==========

The :class:`Screenshot` is used for capturing a digital image of what
is currently visible on the monitor.

The default path for taking screenshot is set in each platform implementation.

Simple Examples
---------------

To get the file path::

    >>> screenshot.file_path
    '/sdcard/test.jpg'

To set the file path::

    >>> screenshot.file_path = '/Users/OSXUser/Pictures/screenshot.png'

To take screenshot::

    >>> from plyer import screenshot
    >>> screenshot.capture()
'''


class Screenshot(object):
    '''
    Screenshot facade.
    '''

    _file_path = ''

    def __init__(self, file_path=None):
        self._file_path = file_path

    def capture(self):
        self._capture()

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, location):
        '''
        Location of the screenshot.
        '''

        self._file_path = location

    # private

    def _capture(self, **kwargs):
        raise NotImplementedError()



'''
Android Compass
---------------------
'''

from plyer.facades import Compass
from jnius import PythonJavaClass, java_method, autoclass, cast
from plyer.platforms.android import activity

Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class MFUSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(MFUSensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_MAGNETIC_FIELD_UNCALIBRATED)
        self.values = [None, None, None, None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:6]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class MagneticFieldSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(MagneticFieldSensorListener, self).__init__()
        self.SensorManager = cast(
            'android.hardware.SensorManager',
            activity.getSystemService(Context.SENSOR_SERVICE)
        )
        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_MAGNETIC_FIELD
        )

        self.values = [None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:3]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        # Maybe, do something in future?
        pass


class AndroidCompass(Compass):
    def __init__(self):
        super(AndroidCompass, self).__init__()
        self.bState = False

    def _enable(self):
        if (not self.bState):
            self.listenerm = MagneticFieldSensorListener()
            self.listenermu = MFUSensorListener()
            self.listenerm.enable()
            self.listenermu.enable()
            self.bState = True

    def _disable(self):
        if (self.bState):
            self.bState = False
            self.listenerm.disable()
            self.listenermu.disable()
            del self.listenerm
            del self.listenermu

    def _get_orientation(self):
        if (self.bState):
            return tuple(self.listenerm.values)
        else:
            return (None, None, None)

    def _get_field_uncalib(self):
        if (self.bState):
            return tuple(self.listenermu.values)
        else:
            return (None, None, None, None, None, None)

    def __del__(self):
        if(self.bState):
            self._disable()
        super(self.__class__, self).__del__()


def instance():
    return AndroidCompass()



'''Bluetooth facade.

Returns the following:

* Bluetooth info

Simple Example
--------------

To get the bluetooth status info::
todo: will be extended to get additional bluetooth info
todo: will be extended to allow bluetooth connections etc.

    >>> from plyer import bluetooth
    >>> bluetooth
    'on' or 'off'

Supported Platforms
-------------------
Android, OS X

'''


class Bluetooth(object):
    '''
    Bluetooth facade.
    '''

    @property
    def info(self):
        '''
        Property that returns the info (currently status) of the bluetooth.
        '''
        return self.get_info()

    def get_info(self):
        return self._get_info()

    # private

    def _get_info(self, **kwargs):
        raise NotImplementedError()



'''
iOS Gravity
-----------

'''

from plyer.facades import Gravity
from pyobjus import autoclass


class iOSGravity(Gravity):

    def __init__(self):
        self.bridge = autoclass('bridge').alloc().init()
        self.bridge.motionManager.setDeviceMotionUpdateInterval_(0.1)

    def _enable(self):
        self.bridge.startDeviceMotion()

    def _disable(self):
        self.bridge.stopDeviceMotion()

    def _get_gravity(self):
        return (
            self.bridge.g_x,
            self.bridge.g_y,
            self.bridge.g_z)


def instance():
    return iOSGravity()



'''
iOS Barometer
-------------
'''

from plyer.facades import Barometer
from pyobjus import autoclass


class iOSBarometer(Barometer):

    def __init__(self):
        super(iOSBarometer, self).__init__()
        self.bridge = autoclass('bridge').alloc().init()

    def _enable(self):
        self.bridge.startRelativeAltitude()

    def _disable(self):
        self.bridge.stopRelativeAltitude()

    def _get_pressure(self):
        '''
        1 kPa = 10 hPa
        '''
        return (
            self.bridge.pressure * 10)


def instance():
    return iOSBarometer()



'''
IOS file chooser
--------------------

This module houses the iOS implementation of the plyer FileChooser.

.. versionadded:: 1.4.4
'''

from plyer.facades import FileChooser
from pyobjus import autoclass, protocol
from pyobjus.dylib_manager import load_framework


load_framework('/System/Library/Frameworks/Photos.framework')


class IOSFileChooser(FileChooser):
    '''
    FileChooser implementation for IOS using
    the built-in file browser via UIImagePickerController.

    .. versionadded:: 1.4.0
    '''

    def __init__(self, *args, **kwargs):
        super(IOSFileChooser, self).__init__(*args, **kwargs)
        self._on_selection = None

    def _file_selection_dialog(self, *args, **kwargs):
        """
        Function called when action is required, A "mode" parameter specifies
        which and is one of "open", "save" or "dir".
        """
        self._on_selection = kwargs["on_selection"]
        if kwargs["mode"] == "open":
            self._open()
        else:
            raise NotImplementedError()

    def _get_picker(self):
        """
        Return an instantiated and configured UIImagePickerController.
        """
        picker = autoclass("UIImagePickerController")
        po = picker.alloc().init()
        po.sourceType = 0
        po.delegate = self
        return po

    def _open(self):
        """
        Launch the native iOS file browser. Upon selection, the
        `imagePickerController_didFinishPickingMediaWithInfo_` delegate is
        called where we close the file browser and handle the result.
        """
        picker = self._get_picker()
        UIApplication = autoclass('UIApplication')
        vc = UIApplication.sharedApplication().keyWindow.rootViewController()
        vc.presentViewController_animated_completion_(picker, True, None)

    @protocol('UIImagePickerControllerDelegate')
    def imagePickerController_didFinishPickingMediaWithInfo_(
            self, image_picker, frozen_dict):
        """
        Delegate which handles the result of the image seletion process.
        """
        image_picker.dismissViewControllerAnimated_completion_(True, None)

        # Note: We need to call this Objective C class as there is currently
        #       no way to call a non-class function via pyobjus. And here,
        #       we have to use the `UIImagePNGRepresentation` to get the png
        #       representation. For this, please ensure you are using an
        #       appropriate version of kivy-ios.
        native_image_picker = autoclass("NativeImagePicker").alloc().init()
        path = native_image_picker.writeToPNG_(frozen_dict)
        self._on_selection([path.UTF8String()])


def instance():
    return IOSFileChooser()



from jnius import autoclass
from jnius import cast
from jnius import java_method
from jnius import PythonJavaClass
from plyer.platforms.android import activity
from plyer.facades import SpatialOrientation

Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class AccelerometerSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(AccelerometerSensorListener, self).__init__()
        self.SensorManager = cast(
            'android.hardware.SensorManager',
            activity.getSystemService(Context.SENSOR_SERVICE)
        )
        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_ACCELEROMETER
        )
        self.values = [None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:3]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class MagnetometerSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(MagnetometerSensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_MAGNETIC_FIELD)
        self.values = [None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:3]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidSpOrientation(SpatialOrientation):

    def __init__(self):
        self.state = False

    def _get_orientation(self):
        if self.state:
            rotation = [0] * 9
            inclination = [0] * 9
            gravity = []
            geomagnetic = []
            gravity = self.listener_a.values
            geomagnetic = self.listener_m.values
            if gravity[0] is not None and geomagnetic[0] is not None:
                ff_state = SensorManager.getRotationMatrix(
                    rotation, inclination,
                    gravity, geomagnetic
                )
                if ff_state:
                    values = [0, 0, 0]
                    values = SensorManager.getOrientation(
                        rotation, values
                    )
                return values

    def _enable_listener(self, **kwargs):
        if not self.state:
            self.listener_a = AccelerometerSensorListener()
            self.listener_m = MagnetometerSensorListener()
            self.listener_a.enable()
            self.listener_m.enable()
            self.state = True

    def _disable_listener(self, **kwargs):
        if self.state:
            self.listener_a.disable()
            self.listener_m.disable()
            self.state = False
            delattr(self, 'listener_a')
            delattr(self, 'listener_m')


def instance():
    return AndroidSpOrientation()



import subprocess
from os.path import join
from plyer.facades import Screenshot
from plyer.utils import whereis_exe
from plyer.platforms.linux.storagepath import LinuxStoragePath


class LinuxScreenshot(Screenshot):
    def __init__(self, file_path=None):
        default_path = join(
            LinuxStoragePath().get_pictures_dir(),
            'screenshot.xwd'
        )
        super(LinuxScreenshot, self).__init__(file_path or default_path)

    def _capture(self):
        # call xwd and redirect bytes from stdout to file
        with open(self.file_path, 'wb') as fle:
            subprocess.call([
                # quiet, full screen root window
                'xwd', '-silent', '-root',
            ], stdout=fle)


def instance():
    if whereis_exe('xwd'):
        return LinuxScreenshot()
    else:
        return Screenshot()



'''
Accelerometer
============

The accelerometer is a motion sensor that detects the change (delta) in
movement relative to the current device orientation, in three dimensions
along the x, y, and z axis.

The :class:`Accelerometer` provides access to public methods to
use accelerometer of your device.

Simple Examples
---------------

To enable accelerometer::

    >>> from plyer import accelerometer
    >>> accelerometer.enable()

To disable acceletometer::

    >>> accelerometer.disable()

To get the acceleration::

    >>> accelerometer.acceleration
    (-10.048464775085449, 6.825869083404541, 7.7260890007019043)

Supported Plaforms
------------------
Android, iOS, OS X, Linux

'''


class Accelerometer(object):
    '''
    Accelerometer facade.
    '''

    @property
    def acceleration(self):
        '''
        Property that returns values of the current acceleration
        sensors, as a (x, y, z) tuple. Returns (None, None, None)
        if no data is currently available.
        '''
        return self.get_acceleration()

    def enable(self):
        '''
        Activate the accelerometer sensor. Throws an error if the
        hardware is not available or not implemented on.
        '''
        self._enable()

    def disable(self):
        '''
        Disable the accelerometer sensor.
        '''
        self._disable()

    def get_acceleration(self):
        return self._get_acceleration()

    # private

    def _enable(self):
        raise NotImplementedError()

    def _disable(self):
        raise NotImplementedError()

    def _get_acceleration(self):
        raise NotImplementedError()



'''
Module of Linux API for plyer.notification.
'''

import warnings
import subprocess
from plyer.facades import Notification
from plyer.utils import whereis_exe


class NotifySendNotification(Notification):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of Linux notification API
    using notify-send binary.
    '''
    def _notify(self, **kwargs):
        subprocess.call([
            "notify-send", kwargs.get('title'), kwargs.get('message')
        ])


class NotifyDbus(Notification):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of Linux notification API
    using dbus library and dbus-python wrapper.
    '''

    def _notify(self, **kwargs):
        # pylint: disable=too-many-locals
        summary = kwargs.get('title', "title")
        body = kwargs.get('message', "body")
        app_name = kwargs.get('app_name', '')
        app_icon = kwargs.get('app_icon', '')
        timeout = kwargs.get('timeout', 10)
        actions = kwargs.get('actions', [])
        hints = kwargs.get('hints', [])
        replaces_id = kwargs.get('replaces_id', 0)

        _bus_name = 'org.freedesktop.Notifications'
        _object_path = '/org/freedesktop/Notifications'
        _interface_name = _bus_name

        import dbus  # pylint: disable=import-error
        session_bus = dbus.SessionBus()
        obj = session_bus.get_object(_bus_name, _object_path)
        interface = dbus.Interface(obj, _interface_name)
        interface.Notify(
            app_name, replaces_id, app_icon,
            summary, body, actions,
            hints, timeout * 1000
        )


def instance():
    '''
    Instance for facade proxy.
    '''
    try:
        import dbus  # pylint: disable=unused-variable,unused-import
        return NotifyDbus()
    except ImportError:
        msg = ("The Python dbus package is not installed.\n"
               "Try installing it with your distribution's package manager, "
               "it is usually called python-dbus or python3-dbus, but you "
               "might have to try dbus-python instead, e.g. when using pip.")
        warnings.warn(msg)

    if whereis_exe('notify-send'):
        return NotifySendNotification()
    warnings.warn("notify-send not found.")
    return Notification()



'''
Windows file chooser
--------------------
'''

from plyer.facades import FileChooser
from win32com.shell.shell import (
    SHBrowseForFolder as browse,
    SHGetPathFromIDList as get_path, SHILCreateFromPath
)
from win32com.shell import shellcon
import win32gui
import win32con
import pywintypes
import pathlib
from os.path import dirname, splitext, join, isdir


class Win32FileChooser(object):
    '''A native implementation of NativeFileChooser using the
    Win32 API on Windows.

    Not Implemented features (all dialogs):
    * preview
    * icon

    Not implemented features (in directory selection only - it's limited
    by Windows itself):
    * preview
    * window-icon

    Known issues:
    * non-existins folders such as: Network, Control Panel, My Computer, Trash,
      Library and likes will raise a COM error. The path does not exist, nor
      a user can open from or save to such path.
    '''

    path = None
    multiple = False
    filters = []
    preview = False
    title = None
    icon = None
    show_hidden = False

    def __init__(self, *args, **kwargs):
        self._handle_selection = kwargs.pop(
            'on_selection', self._handle_selection
        )

        # Simulate Kivy's behavior
        for i in kwargs:
            setattr(self, i, kwargs[i])

    @staticmethod
    def _handle_selection(selection):  # pylint: disable=method-hidden
        '''
        Dummy placeholder for returning selection from chooser.
        '''
        return selection

    def run(self):
        self.selection = []
        try:
            if self.mode != "dir":
                args = {}

                if self.path:
                    if isdir(self.path):
                        args["InitialDir"] = self.path
                    else:
                        args["InitialDir"] = dirname(self.path)
                        _, ext = splitext(self.path)
                        args["File"] = self.path
                        args["DefExt"] = ext and ext[1:]  # no period

                args["Title"] = self.title if self.title else "Pick a file..."
                args["CustomFilter"] = 'Other file types\x00*.*\x00'
                args["FilterIndex"] = 1

                # e.g. open_file(filters=['*.txt', '*.py'])
                filters = ""
                for f in self.filters:
                    if type(f) == str:
                        filters += (f + "\x00") * 2
                    else:
                        filters += f[0] + "\x00" + ";".join(f[1:]) + "\x00"
                args["Filter"] = filters

                flags = win32con.OFN_OVERWRITEPROMPT
                flags |= win32con.OFN_HIDEREADONLY

                if self.multiple:
                    flags |= win32con.OFN_ALLOWMULTISELECT
                    flags |= win32con.OFN_EXPLORER
                if self.show_hidden:
                    flags |= win32con.OFN_FORCESHOWHIDDEN

                args["Flags"] = flags

                try:
                    if self.mode == "open":
                        self.fname, _, _ = win32gui.GetOpenFileNameW(**args)
                    elif self.mode == "save":
                        self.fname, _, _ = win32gui.GetSaveFileNameW(**args)
                except pywintypes.error as e:
                    # if canceled, it's not really an error
                    if not e.winerror:
                        self._handle_selection(self.selection)
                        return self.selection
                    raise

                if self.fname:
                    if self.multiple:
                        seq = str(self.fname).split("\x00")
                        if len(seq) > 1:
                            dir_n, base_n = seq[0], seq[1:]
                            self.selection = [
                                join(dir_n, i) for i in base_n
                            ]
                        else:
                            self.selection = seq
                    else:
                        self.selection = str(self.fname).split("\x00")

            else:  # dir mode
                BIF_EDITBOX = shellcon.BIF_EDITBOX
                BIF_NEWDIALOGSTYLE = 0x00000040
                # From http://goo.gl/UDqCqo
                pidl, name, images = browse(  # pylint: disable=unused-variable
                    win32gui.GetDesktopWindow(),
                    None,
                    self.title if self.title else "Pick a folder...",
                    BIF_NEWDIALOGSTYLE | BIF_EDITBOX, None, None
                )

                # pidl is None when nothing is selected
                # and e.g. the dialog is closed afterwards with Cancel
                if pidl:
                    self.selection = [str(get_path(pidl).decode('utf-8'))]

        except (RuntimeError, pywintypes.error, Exception):
            # ALWAYS! let user know what happened
            import traceback
            traceback.print_exc()
        self._handle_selection(self.selection)
        return self.selection


class WinFileChooser(FileChooser):
    '''FileChooser implementation for Windows, using win3all.
    '''

    def _file_selection_dialog(self, **kwargs):
        return Win32FileChooser(**kwargs).run()


def instance():
    return WinFileChooser()



'''
Number of Processors
=======

Simple Example
---------------
To get battery status::
    >>> from plyer import processors
    >>> processors.status
    {'Number_of_Processors': '4'}
Supported Platforms
-------------------
Linux
'''


class Processors(object):
    '''
    Number of Processors info facade.
    '''

    @property
    def status(self):
        '''
        Property that contains a dict with the following fields:
             * **Number_of_Processors** *(int)*: Number of Processors in
             the system
            .. warning::
                If any of the fields is not readable, it is set as
                None.
        '''
        return self.get_state()

    def get_state(self):
        return self._get_state()

    # private

    def _get_state(self):
        raise NotImplementedError()



class Keystore(object):
    '''
    Keyring facade

    .. versionadded:: x.x.x

    '''

    def set_key(self, servicename, key, value, **kwargs):
        self._set_key(servicename, key, value, **kwargs)

    def _set_key(self, servicename, key, value, **kwargs):
        raise NotImplementedError()

    def get_key(self, servicename, key, **kwargs):
        return self._get_key(servicename, key)

    def _get_key(self, servicename, key, **kwargs):
        raise NotImplementedError()



'''
IOS Sms
----------
'''

from plyer.facades import Sms
from pyobjus import autoclass, objc_str
from pyobjus.dylib_manager import load_framework

NSURL = autoclass('NSURL')
NSString = autoclass('NSString')
UIApplication = autoclass('UIApplication')
load_framework('/System/Library/Frameworks/MessageUI.framework')


class IOSSms(Sms):

    def _send(self, **kwargs):
        '''
        This method provides sending messages to recipients.

        Expects 2 parameters in kwargs:
            - recipient: String type
            - message: String type

        Opens a mesage interface with recipient and message information.
        '''
        recipient = kwargs.get('recipient')
        message = kwargs.get('message')
        url = "sms:"
        if recipient:
            # Apple has not supported multiple recipients yet.
            url += str(recipient)
        if message:
            # Apple has to supported it yet.
            pass

        nsurl = NSURL.alloc().initWithString_(objc_str(url))
        UIApplication.sharedApplication().openURL_(nsurl)


def instance():
    return IOSSms()



'''
https://docs.microsoft.com/en-us/windows/desktop/api/wingdi/nf-wingdi-bitblt
https://www.bugs.python.org/issue33656
'''

from os import getpid
from os.path import join
from ctypes import windll, c_int, addressof

from win32gui import GetDesktopWindow, GetWindowDC
from win32api import GetSystemMetrics
from win32ui import CreateDCFromHandle, CreateBitmap
from win32con import (
    SM_CXVIRTUALSCREEN,
    SM_CYVIRTUALSCREEN,
    SM_XVIRTUALSCREEN,
    SM_YVIRTUALSCREEN,
    SRCCOPY
)

from plyer.facades import Screenshot
from plyer.platforms.win.storagepath import WinStoragePath


class WinScreenshot(Screenshot):
    def __init__(self, file_path=None):
        default_path = join(
            WinStoragePath().get_pictures_dir(), 'screenshot.bmp'
        )
        super(WinScreenshot, self).__init__(file_path or default_path)

    def _set_dpi_aware(self, value):
        try:
            windll.shcore.SetProcessDpiAwareness(value)
        except (AttributeError, OSError):
            print('Could not set DPI awareness.')

    def _dpi_aware(self):
        # make backup of DPI awareness value in case a user does not want
        # to use it, otherwise we'll cripple user's runtime
        try:
            process_handle = windll.kernel32.OpenProcess(
                0,      # no permissions
                False,  # bInheritHandle
                getpid()
            )
            aware = c_int()
            windll.shcore.GetProcessDpiAwareness(
                process_handle,
                addressof(aware)
            )
        finally:
            # always close the handle!
            windll.kernel32.CloseHandle(process_handle)
        return bool(aware)

    def _capture(self):
        # make sure the process is DPI aware, otherwise the image
        # is only a part of the full monitor content
        # (necessary only on Win 8.1+)
        old_awareness = self._dpi_aware()
        self._set_dpi_aware(True)

        # get width and height of current monitor
        width = GetSystemMetrics(SM_CXVIRTUALSCREEN)
        height = GetSystemMetrics(SM_CYVIRTUALSCREEN)

        # get 'desktop' window handle
        handle_desktop = GetDesktopWindow()

        # get window graphic context handle
        handle_context = GetWindowDC(handle_desktop)

        # create device context
        dev_ctx = CreateDCFromHandle(handle_context)

        # create destination for original device context
        dest_ctx = dev_ctx.CreateCompatibleDC()

        # create bitmap compatible with desktop window device
        bmp = CreateBitmap()
        bmp.CreateCompatibleBitmap(dev_ctx, width, height)

        # select bitmap into destination device
        dest_ctx.SelectObject(bmp)

        # populate selected bitmap in destination device
        dest_ctx.BitBlt(
            (0, 0),           # start point (x, y)
            (width, height),  # size of rectangle
            dev_ctx, (        # source device
                GetSystemMetrics(SM_XVIRTUALSCREEN),
                GetSystemMetrics(SM_YVIRTUALSCREEN)
            ),                # source rectangle, can be different monitor
            SRCCOPY           # copy directly without filters
        )

        # save bitmap to file
        bmp.SaveBitmapFile(dest_ctx, self.file_path)

        # return to the original state
        self._set_dpi_aware(old_awareness)


def instance():
    return WinScreenshot()



'''
Plyer
=====

'''

__all__ = (
    'accelerometer', 'audio', 'barometer', 'battery', 'bluetooth',
    'brightness', 'call', 'camera', 'compass', 'cpu', 'email', 'filechooser',
    'flash', 'gps', 'gravity', 'gyroscope', 'humidity', 'irblaster',
    'keystore', 'light', 'notification', 'orientation', 'processors',
    'proximity', 'screenshot', 'sms', 'spatialorientation', 'storagepath',
    'stt', 'temperature', 'tts', 'uniqueid', 'vibrator', 'wifi'
)

__version__ = '1.4.3'


from plyer import facades
from plyer.utils import Proxy

# pylint: disable=invalid-name
#: Accelerometer proxy to :class:`plyer.facades.Accelerometer`
accelerometer = Proxy('accelerometer', facades.Accelerometer)

#: Keyring proxy to :class::`plyer.facades.Keyring`
keystore = Proxy('keystore', facades.Keystore)

#: Audio proxy to :class:`plyer.facades.Audio`
audio = Proxy('audio', facades.Audio)

#: Barometer proxy to :class:`plyer.facades.Barometer`
barometer = Proxy('barometer', facades.Barometer)

#: Battery proxy to :class:`plyer.facades.Battery`
battery = Proxy('battery', facades.Battery)

#: Call proxy to  :class `plyer.facades.Call`
call = Proxy('call', facades.Call)

#: Compass proxy to :class:`plyer.facades.Compass`
compass = Proxy('compass', facades.Compass)

#: Camera proxy to :class:`plyer.facades.Camera`
camera = Proxy('camera', facades.Camera)

#: Email proxy to :class:`plyer.facades.Email`
email = Proxy('email', facades.Email)

#: FileChooser proxy to :class:`plyer.facades.FileChooser`
filechooser = Proxy('filechooser', facades.FileChooser)

#: GPS proxy to :class:`plyer.facades.GPS`
gps = Proxy('gps', facades.GPS)

#: Gravity proxy to :class:`plyer.facades.Gravity`
gravity = Proxy('gravity', facades.Gravity)

#: Gyroscope proxy to :class:`plyer.facades.Gyroscope`
gyroscope = Proxy('gyroscope', facades.Gyroscope)

#: IrBlaster proxy to :class:`plyer.facades.IrBlaster`
irblaster = Proxy('irblaster', facades.IrBlaster)

#: Light proxy to :class:`plyer.facades.Light`
light = Proxy('light', facades.Light)

#: Orientation proxy to :class:`plyer.facades.Orientation`
orientation = Proxy('orientation', facades.Orientation)

#: Notification proxy to :class:`plyer.facades.Notification`
notification = Proxy('notification', facades.Notification)

#: Proximity proxy to :class:`plyer.facades.Proximity`
proximity = Proxy('proximity', facades.Proximity)

#: Sms proxy to :class:`plyer.facades.Sms`
sms = Proxy('sms', facades.Sms)

#: Speech proxy to :class:`plyer.facades.STT`
stt = Proxy('stt', facades.STT)

#: TTS proxy to :class:`plyer.facades.TTS`
tts = Proxy('tts', facades.TTS)

#: UniqueID proxy to :class:`plyer.facades.UniqueID`
uniqueid = Proxy('uniqueid', facades.UniqueID)

#: Vibrator proxy to :class:`plyer.facades.Vibrator`
vibrator = Proxy('vibrator', facades.Vibrator)

#: Flash proxy to :class:`plyer.facades.Flash`
flash = Proxy('flash', facades.Flash)

#: Wifi proxy to :class:`plyer.facades.Wifi`
wifi = Proxy('wifi', facades.Wifi)

#: Temperature proxy to :class:`plyer.facades.Temperature`
temperature = Proxy('temperature', facades.Temperature)

#: Humidity proxy to :class:`plyer.facades.Humidity`
humidity = Proxy('humidity', facades.Humidity)
#: SpatialOrientation proxy to :class:`plyer.facades.SpatialOrientation`
spatialorientation = Proxy('spatialorientation', facades.SpatialOrientation)

#: Brightness proxy to :class:`plyer.facades.Brightness`
brightness = Proxy('brightness', facades.Brightness)

#: StoragePath proxy to :class:`plyer.facades.StoragePath`
storagepath = Proxy('storagepath', facades.StoragePath)

#: Bluetooth proxy to :class:`plyer.facades.Bluetooth`
bluetooth = Proxy('bluetooth', facades.Bluetooth)

#: Processors proxy to :class:`plyer.facades.Processors`
processors = Proxy('processors', facades.Processors)

#: Processors proxy to :class:`plyer.facades.CPU`
cpu = Proxy('cpu', facades.CPU)

#: Screenshot proxy to :class:`plyer.facades.Screenshot`
screenshot = Proxy('screenshot', facades.Screenshot)



'''
Module of iOS API for plyer.uniqueid.
'''

from pyobjus import autoclass  # pylint: disable=import-error
from pyobjus.dylib_manager import load_framework  # pylint:disable=import-error
from plyer.facades import UniqueID

load_framework('/System/Library/Frameworks/UIKit.framework')
UIDevice = autoclass('UIDevice')


class IOSUniqueID(UniqueID):
    '''
    Implementation of iOS uniqueid API.
    '''

    def _get_uid(self):
        uuid = UIDevice.currentDevice().identifierForVendor.UUIDString()
        return uuid.UTF8String()


def instance():
    '''
    Instance for facade proxy.
    '''
    return IOSUniqueID()



from subprocess import Popen, PIPE
from plyer.facades import Processors
from plyer.utils import whereis_exe

from os import environ


class LinuxProcessors(Processors):
    def _get_state(self):
        old_lang = environ.get('LANG')
        environ['LANG'] = 'C'

        status = {"Number_of_Processors": None}

        dev = "--all"
        nproc_process = Popen(
            ["nproc", dev],
            stdout=PIPE
        )
        output = nproc_process.communicate()[0]

        environ['LANG'] = old_lang

        if not output:
            return status

        status['Number_of_Processors'] = output.rstrip()

        return status


def instance():
    import sys
    if whereis_exe('nproc'):
        return LinuxProcessors()
    sys.stderr.write("nproc not found.")
    return Processors()



from os.path import join

from pyobjus import autoclass
from pyobjus.dylib_manager import INCLUDE, load_framework

from plyer.facades import Audio
from plyer.platforms.macosx.storagepath import OSXStoragePath

load_framework(INCLUDE.Foundation)
load_framework(INCLUDE.AVFoundation)

AVAudioPlayer = autoclass("AVAudioPlayer")
AVAudioRecorder = autoclass("AVAudioRecorder")
AVAudioFormat = autoclass("AVAudioFormat")
NSString = autoclass('NSString')
NSURL = autoclass('NSURL')
NSError = autoclass('NSError').alloc()


class OSXAudio(Audio):
    def __init__(self, file_path=None):
        default_path = join(
            OSXStoragePath().get_music_dir(),
            'audio.wav'
        )
        super(OSXAudio, self).__init__(file_path or default_path)

        self._recorder = None
        self._player = None
        self._current_file = None

    def _start(self):
        # Conversion of Python file path string to Objective-C NSString
        file_path_NSString = NSString.alloc()
        file_path_NSString = file_path_NSString.initWithUTF8String_(
            self._file_path
        )

        # Definition of Objective-C NSURL object for the output record file
        # specified by NSString file path
        file_NSURL = NSURL.alloc()
        file_NSURL = file_NSURL.initWithString_(file_path_NSString)

        # Internal audio file format specification
        af = AVAudioFormat.alloc()
        af = af.initWithCommonFormat_sampleRate_channels_interleaved_(
            1, 44100.0, 2, True
        )

        # Audio recorder instance initialization with specified file NSURL
        # and audio file format
        self._recorder = AVAudioRecorder.alloc()
        self._recorder = self._recorder.initWithURL_format_error_(
            file_NSURL, af, NSError
        )

        if not self._recorder:
            raise Exception(NSError.code, NSError.domain)

        self._recorder.record()

        # Setting the currently recorded file as current file
        # for using it as a parameter in audio player
        self._current_file = file_NSURL

    def _stop(self):
        if self._recorder:
            self._recorder.stop()
            self._recorder = None

        if self._player:
            self._player.stop()
            self._player = None

    def _play(self):
        # Audio player instance initialization with the file NSURL
        # of the last recorded audio file
        self._player = AVAudioPlayer.alloc()
        self._player = self._player.initWithContentsOfURL_error_(
            self._current_file, NSError
        )

        if not self._player:
            raise Exception(NSError.code, NSError.domain)

        self._player.play()


def instance():
    return OSXAudio()



'''
TTS
====

The :class:`TTS` provides provides access to public methods to
use text to speech of your device.

Simple Examples
---------------

To speak::

    >>> from plyer import tts
    >>> tts.speak(message=message)

Supported Platforms
-------------------
Android, iOS, Windows, OS X, Linux

'''


class TTS(object):
    '''
    TextToSpeech facade.
    '''

    def speak(self, message=''):
        '''Use text to speech capabilities to speak the message.

        :param message: What to speak
        :type message: str
        '''
        self._speak(message=message)

    # private

    def _speak(self, **kwargs):
        raise NotImplementedError()



try:
    import keyring
except ImportError:
    raise NotImplemented()
from plyer.facades import Keystore


class OSXKeystore(Keystore):

    def _set_key(self, servicename, key, value, **kwargs):
        keyring.set_password(servicename, key, value)

    def _get_key(self, servicename, key, **kwargs):
        return keyring.get_password(servicename, key)


def instance():
    return OSXKeystore()



'''
Module of MacOS API for plyer.battery.
'''

from os import environ
from subprocess import Popen, PIPE
from plyer.facades import Battery
from plyer.utils import whereis_exe


class OSXBattery(Battery):
    '''
    Implementation of MacOS battery API.
    '''

    def _get_state(self):
        old_lang = environ.get('LANG', '')
        environ['LANG'] = 'C'

        status = {"isCharging": None, "percentage": None}

        ioreg_process = Popen(
            ["ioreg", "-rc", "AppleSmartBattery"],
            stdout=PIPE
        )
        output = ioreg_process.communicate()[0]

        environ['LANG'] = old_lang

        if not output:
            return status

        is_charging = max_capacity = current_capacity = None
        for line in output.decode('utf-8').splitlines():
            if 'IsCharging' in line:
                is_charging = line.rpartition('=')[-1].strip()
            if 'MaxCapacity' in line:
                max_capacity = float(line.rpartition('=')[-1].strip())
            if 'CurrentCapacity' in line:
                current_capacity = float(line.rpartition('=')[-1].strip())

        if is_charging:
            status['isCharging'] = is_charging == "Yes"

        if current_capacity and max_capacity:
            status['percentage'] = 100.0 * current_capacity / max_capacity

        return status


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('ioreg'):
        return OSXBattery()
    sys.stderr.write("ioreg not found.")
    return Battery()



'''
Notification
============

The :class:`Notification` provides access to public methods to create
notifications.

Simple Examples
---------------

To send notification::

    >>> from plyer import notification
    >>> title = 'plyer'
    >>> message = 'This is an example.'
    >>> notification.notify(title=title, message=message)

Android toast notification::

    >>> from plyer import notification
    >>> notification.notify(message='hello', toast=True)

Android simple notification::

    >>> from plyer import notification
    >>> notification.notify(message='hello', toast=True)

Notification with custom icon::

    >>> from plyer import notification
    >>> notification.notify(title='title', message='hello', app_icon=<path>)

.. versionadded:: 1.0.0

.. versionadded:: 1.4.0
   Add implementation of primitive Android popup-like notification (toast)

.. versionchanged:: 1.4.0
   Android implementation now supports custom icons for notifications.
'''


class Notification(object):
    # pylint: disable=too-few-public-methods
    '''
    Notification facade.
    '''

    def notify(self, title='', message='', app_name='', app_icon='',
               timeout=10, ticker='', toast=False):
        # pylint: disable=too-many-arguments
        '''
        Send a notification.

        :param title: Title of the notification
        :param message: Message of the notification
        :param app_name: Name of the app launching this notification
        :param app_icon: Icon to be displayed along with the message
        :param timeout: time to display the message for, defaults to 10
        :param ticker: text to display on status bar as the notification
                       arrives
        :param toast: simple Android message instead of full notification
        :type title: str
        :type message: str
        :type app_name: str
        :type app_icon: str
        :type timeout: int
        :type ticker: str
        :type toast: bool

        .. note::
           When called on Windows, ``app_icon`` has to be a path to
           a file in .ICO format.

        .. versionadded:: 1.0.0

        .. versionchanged:: 1.4.0
           Add 'toast' keyword argument
        '''

        self._notify(
            title=title, message=message,
            app_icon=app_icon, app_name=app_name,
            timeout=timeout, ticker=ticker, toast=toast
        )

    # private

    def _notify(self, **kwargs):
        raise NotImplementedError("No usable implementation found!")



import subprocess
from os.path import join
from plyer.facades import Screenshot
from plyer.utils import whereis_exe
from plyer.platforms.macosx.storagepath import OSXStoragePath


class OSXScreenshot(Screenshot):
    def __init__(self, file_path=None):
        default_path = join(
            OSXStoragePath().get_pictures_dir().replace('file://', ''),
            'screenshot.png'
        )
        super(OSXScreenshot, self).__init__(file_path or default_path)

    def _capture(self):
        subprocess.call([
            'screencapture',
            self.file_path
        ])


def instance():
    if whereis_exe('screencapture'):
        return OSXScreenshot()
    else:
        return Screenshot()



'''
Brightness
==========

This API helps you to control the brightness of your primary display screen.

The :class:`Brightness` provides access to public methods to control the
brightness of screen.

NOTE:: For Android, make sure to add permission, WRITE_SETTINGS

Simple Examples
---------------
To know the current brightness level of device::

    >>> from plyer import brightness
    >>> brightness.current_level()

To set the brightness level to half of maximum::

    >>> from plyer import brightness
    >>> brightness.set_level(50)

Supported Platforms
-------------------
Android, iOS, Linux
'''


class Brightness(object):
    '''
    Brightness facade.
    '''

    def current_level(self):
        '''
        Know the current level of device's brightness.
        '''
        return self._current_level()

    def set_level(self, level):
        '''
        Adjust the brightness of the screen.
        Minimum brightnesss level:: 1
        Maximum brightness level:: 100

        :param level: New level of brightness between 1 and 100
        :type level: int
        '''
        return self._set_level(level)

    # private

    def _set_level(self, level):
        raise NotImplementedError()

    def _current_level(self):
        raise NotImplementedError()



'''
Module of MacOS API for plyer.notification.
'''

from plyer.facades import Notification

try:
    import pyobjus  # pylint: disable=unused-import
except ImportError:
    raise Exception(
        "pyobjus missing, please install it. "
        "`python -m pip install git+http://github.com/kivy/pyobjus/`"
    )

from pyobjus import (  # pylint: disable=import-error
    autoclass, protocol, objc_str, ObjcBOOL
)
from pyobjus.dylib_manager import (  # pylint: disable=import-error
    load_framework, INCLUDE
)

load_framework(INCLUDE.AppKit)
load_framework(INCLUDE.Foundation)

NSUserNotification = autoclass('NSUserNotification')
NSUserNotificationCenter = autoclass('NSUserNotificationCenter')


class OSXNotification(Notification):
    '''
    Implementation of MacOS notification API.
    '''

    def _notify(self, **kwargs):
        title = kwargs.get('title', '')
        message = kwargs.get('message', '')
        app_name = kwargs.get('app_name', '')
        # app_icon, timeout, ticker are not supported (yet)

        notification = NSUserNotification.alloc().init()
        notification.setTitle_(objc_str(title))
        notification.setSubtitle_(objc_str(app_name))
        notification.setInformativeText_(objc_str(message))

        usrnotifctr = NSUserNotificationCenter.defaultUserNotificationCenter()
        usrnotifctr.setDelegate_(self)
        usrnotifctr.deliverNotification_(notification)

    @protocol('NSUserNotificationCenterDelegate')
    def userNotificationCenter_shouldPresentNotification_(
            self, center, notification):
        # pylint: disable=invalid-name,missing-docstring
        # pylint: disable=unused-argument,no-self-use
        return ObjcBOOL(True)


def instance():
    '''
    Instance for facade proxy.
    '''
    return OSXNotification()



import plyer.platforms.win.libs.wifi_defs as wifi_lib
from plyer.facades import Wifi


class WindowWifi(Wifi):

    names = {}

    def _is_enabled(self):
        '''
        TODO: Implement this in future
        Couldn't find a nice implementation for this although
        NetworkInformation class could be used but ctypes doesn't supports
        class yet. It should look something like this.

        for item in NetworkInformation.getConnectionProfiles():
            if item.IsWlanConnectionProfile:
                adapter_id = item.NetworkAdapter.NetworkAdapterId
        for item in NetworkInformation.GetLanIdentifiers():
            if item.NetworkAdapterId == adapter_id:
                is_wifi_enabled = True
        return True/False

        Returning True for now to make it work.
        '''
        return True

    def _get_network_info(self, name):
        '''
        Returns all the network information.
        '''
        return wifi_lib.get_network_info(name)

    def _start_scanning(self):
        '''
        Starts scanning for available Wi-Fi networks and returns the available,
        devices.
        '''
        if self._is_enabled():
            self.names = wifi_lib.start_scanning()
        else:
            raise Exception('Wifi not Enabled.')

    def _get_available_wifi(self):
        '''
        Returns the name of available networks.
        '''
        return wifi_lib.get_available_wifi()

    def _connect(self, network, parameters):
        '''
        Expects 2 parameters:
          - name/ssid of the network.
          - parameters: dict type
            - connection_mode:
              `https://msdn.microsoft.com/en-us/library/windows/desktop/
               ms706844(v=vs.85).aspx`

              :between range [0, 5]
              - wlan_connection_mode_profile,
                wlan_connection_mode_temporary_profile,
                wlan_connection_mode_discovery_secure,
                wlan_connection_mode_discovery_unsecure,
                wlan_connection_mode_auto,
                wlan_connection_mode_invalid.
            - profile:
                if wlanConnectionMode = wlan_connection_mode_profile
                then profile = ssid
                if wlanConnectionMode = wlan_connection_mode_temporary_profile
                then profile = XML representation of the profile used for the
                               connection
                if wlanConnectionMode = wlan_connection_mode_discovery_secure
                                        or
                                        wlan_connection_mode_discovery_unsecure
                then profile = None
            - ssid: optional (as network name and ssid are same)
            - bssidList
              `https://msdn.microsoft.com/en-us/library/windows/desktop/
              ms705996(v=vs.85).aspx`
              - Header
                structure that contains the type, version, and, size
                information
                of an NDIS structure.
                - Type: NDSI_OBJECT_TYPE_DEFAULT
                - Revision: DOT11_BSSID_LIST_REVISION_1
                - Size: sizeof(DOT11_BSSID_LIST)
              - uNumOfEntries
                The number of entries in this structure.
              - uTotalNumOfEntries
                The total number of entries supported.
              - BSSIDs
                `https://msdn.microsoft.com/en-us/library/windows/desktop/
                bb427397(v=vs.85).aspx`
                A list of BSS identifiers.
            - bssType
              `https://msdn.microsoft.com/en-us/library/windows/desktop/
              ms706001(v=vs.85).aspx`
              Constants:
              dot11_BSS_type_infrastructure  = 1,
              dot11_BSS_type_independent     = 2,
              dot11_BSS_type_any             = 3
            - flags
              Constant: WLAN_CONNECTION_HIDDEN_NETWORK  value: 0x00000001
              Constant: WLAN_CONNECTION_ADHOC_JOIN_ONLY value: 0x00000002
              Constant: WLAN_CONNECTION_IGNORE_PRIVACY_BIT value: 0x00000004
              Constant: WLAN_CONNECTION_EAPOL_PASSTHROUGH value: 0x00000008
            - password
        '''
        wifi_lib.connect(network, parameters)
        return

    def _disconnect(self):
        '''
        Disconnect from network.
        '''
        wifi_lib.disconnect()
        return


def instance():
    return WindowWifi()



'''
Compass
=======

The :class:`Compass` provides access to public methods to use compass of your
device.

Simple Examples
---------------

To enable compass::

    >>> from plyer import compass
    >>> compass.enable()

To disable compass::

    >>> compass.disable()

To get the field::

    >>> compass.field()
    (-23.721826553344727, -5.7114701271057129, -36.749668121337891)

To get the uncalibrated field along with iron bias estimation::

    >>> compass.field_uncalib()
    (a,b,c,x,y,z)
    # a,b,c denote the Geomagnetic field strength
    # (without hard iron calibration) along the three axes.
    # x,y,z denote the Iron bias estimation along the three axes.

Supported Platforms
-------------------
Android, iOS

'''


class Compass(object):
    '''Compass facade.

    .. versionadded:: 1.2.0
    '''

    @property
    def orientation(self):
        '''
        WARNING:: This property is deprecated after API level 8.
        Use `compass.field` instead.

        Property that returns values of the current compass
        (magnetic field) sensors, as a (x, y, z) tuple.
        Returns (None, None, None) if no data is currently available.
        '''
        return self.get_orientation()

    @property
    def field(self):
        '''
        .. versionadded:: 1.3.1

        Property that returns values of the current compass
        (magnetic field) sensors, as a (x, y, z) tuple.
        Returns (None, None, None) if no data is currently available.
        '''
        return self.get_orientation()

    @property
    def field_uncalib(self):
        '''
        .. versionadded:: 1.3.1

        Property that returns the current value of Uncalibrated Magnetic Field
        (without hard iron calibration) along with the iron bias estimation
        along the three axes.
        '''
        return self.get_field_uncalib()

    def enable(self):
        '''
        Activate the compass sensor.
        '''
        self._enable()

    def disable(self):
        '''
        Disable the compass sensor.
        '''
        self._disable()

    def get_orientation(self):
        return self._get_orientation()

    def get_field_uncalib(self):
        '''
        .. versionadded:: 1.3.1
        '''
        return self._get_field_uncalib()

    # private

    def _enable(self):
        raise NotImplementedError()

    def _disable(self):
        raise NotImplementedError()

    def _get_orientation(self):
        raise NotImplementedError()

    def _get_field_uncalib(self):
        raise NotImplementedError()



'''
.. note::
   This facade depends on `nmcli` (Network Manager command line tool).
   It's found in most of the popular GNU/Linux distributions. Support for other
   backends is not provided yet.
'''

from subprocess import Popen, PIPE, call
from plyer.facades import Wifi
from plyer.utils import whereis_exe, deprecated


class NMCLIWifi(Wifi):
    '''
    .. versionadded:: 1.4.0
    '''

    def __init__(self, *args, **kwargs):
        '''
        .. versionadded:: 1.4.0
        '''

        super(NMCLIWifi, self).__init__(*args, **kwargs)
        self.names = {}

    @property
    def interfaces(self):
        '''
        Get all the available interfaces for WiFi.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self._is_enabled():
            self._enable()

        # fetch the devices
        proc = Popen([
            'nmcli', '--terse',
            '--fields', 'DEVICE,TYPE',
            'device'
        ], stdout=PIPE)
        lines = proc.communicate()[0].decode('utf-8').splitlines()

        # filter devices by type
        interfaces = []
        for line in lines:
            # bad escape from nmcli's side :<
            line = line.replace('\\:', '$$')
            device, dtype = line.split(':')
            if dtype != 'wifi':
                continue
            interfaces.append(device.replace('$$', ':'))

        # return wifi interfaces
        return interfaces

    def _is_enabled(self):
        '''
        Return the status of WiFi device.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        output = Popen(
            ["nmcli", "radio", "wifi"],
            stdout=PIPE
        ).communicate()[0].decode('utf-8')

        if output.split()[0] == 'enabled':
            return True
        return False

    def _is_connected(self, interface=None):
        '''
        Return whether a specified interface is connected to a WiFi network.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self._is_enabled():
            self._enable()
        if not interface:
            interface = self.interfaces[0]

        # fetch all devices
        proc = Popen([
            'nmcli', '--terse',
            '--fields', 'DEVICE,TYPE,STATE',
            'device'
        ], stdout=PIPE)
        lines = proc.communicate()[0].decode('utf-8').splitlines()

        # filter by wifi type and interface
        connected = False
        for line in lines:
            line = line.replace('\\:', '$$')
            device, dtype, state = line.split(':')
            device = device.replace('$$', ':')
            if dtype != 'wifi':
                continue

            if device != interface:
                continue

            if state == 'connected':
                connected = True

        return connected

    def _start_scanning(self, interface=None):
        '''
        Start scanning for available Wi-Fi networks
        for the specified interface.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self._is_enabled():
            self._enable()
        if not interface:
            interface = self.interfaces[0]

        # force rescan for fresh data
        call(['nmcli', 'device', 'wifi', 'rescan', 'ifname', interface])

        # get properties
        fields = [
            'SSID', 'BSSID', 'MODE', 'CHAN', 'FREQ',
            'BARS', 'RATE', 'SIGNAL', 'SECURITY'
        ]

        # fetch all networks for interface
        output = Popen([
            'nmcli', '--terse',
            '--fields', ','.join(fields),
            'device', 'wifi', 'list', 'ifname', interface
        ], stdout=PIPE).communicate()[0].decode('utf-8')

        # parse output
        for line in output.splitlines():
            line = line.replace('\\:', '$$')
            row = {
                field: value
                for field, value in zip(fields, line.split(':'))
            }

            row['BSSID'] = row['BSSID'].replace('$$', ':')
            self.names[row['SSID']] = row

    def _get_network_info(self, name):
        '''
        Get all the network information by network's name (SSID).

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self.names:
            self._start_scanning()

        ret_list = {}
        ret_list['ssid'] = self.names[name]['SSID']
        ret_list['signal'] = self.names[name]['SIGNAL']

        bars = len(self.names[name]['BARS'])
        ret_list['quality'] = '{}/100'.format(bars / 5.0 * 100)
        ret_list['frequency'] = self.names[name]['FREQ']
        ret_list['bitrates'] = self.names[name]['RATE']

        # wpa1, wpa2, wpa1 wpa2, wep, (none), perhaps something else
        security = self.names[name]['SECURITY'].lower()
        ret_list['encrypted'] = True
        if 'wpa2' in security:
            # wpa2, wpa2+wpa1
            ret_list['encryption_type'] = 'wpa2'
        elif 'wpa' in security:
            ret_list['encryption_type'] = 'wpa'
        elif 'wep' in security:
            ret_list['encryption_type'] = 'wep'
        elif 'none' in security:
            ret_list['encrypted'] = False
            ret_list['encryption_type'] = 'none'
        else:
            ret_list['encryption_type'] = security

        ret_list['channel'] = int(self.names[name]['CHAN'])
        ret_list['address'] = self.names[name]['BSSID']
        ret_list['mode'] = self.names[name]['MODE']
        return ret_list

    def _get_available_wifi(self):
        '''
        Return the names of all found networks.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self.names:
            self._start_scanning()
        return list(self.names.keys())

    def _connect(self, network, parameters, interface=None):
        '''
        Connect a specific interface to a WiFi network.

        Expects 2 parameters:
            - SSID of the network
            - parameters: dict
                - password: string or None

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        self._enable()
        if not interface:
            interface = self.interfaces[0]

        password = parameters.get('password')
        command = [
            'nmcli', 'device', 'wifi', 'connect', network,
            'ifname', interface
        ]
        if password:
            command += ['password', password]
        call(command)

    def _disconnect(self, interface=None):
        '''
        Disconnect a specific interface from a WiFi network.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        if not self._is_enabled():
            return

        if not interface:
            interface = self.interfaces[0]

        if self._nmcli_version() >= (1, 2, 6):
            call(['nmcli', 'device', 'disconnect', interface])
        else:
            call(['nmcli', 'nm', 'enable', 'false'])

    def _enable(self):
        '''
        Turn WiFi device on.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        call(['nmcli', 'radio', 'wifi', 'on'])

    def _disable(self):
        '''
        Turn WiFi device off.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        call(['nmcli', 'radio', 'wifi', 'off'])

    def _nmcli_version(self):
        '''
        Get nmcli version to prevent executing deprecated commands.

        .. versionadded:: 1.4.0
           Tested with nmcli 1.2.6.
        '''
        version = Popen(['nmcli', '-v'], stdout=PIPE)
        version = version.communicate()[0].decode('utf-8')
        while version and not version[0].isdigit():
            version = version[1:]
        return tuple(map(int, (version.split('.'))))


@deprecated
class LinuxWifi(Wifi):
    '''
    .. versionadded:: 1.2.5
    '''

    def __init__(self, *args, **kwargs):
        '''
        .. versionadded:: 1.4.0
        '''

        super(LinuxWifi, self).__init__(*args, **kwargs)
        self.names = {}

    @property
    def interfaces(self):
        '''
        .. versionadded:: 1.4.0
        '''

        proc = Popen([
            'nmcli', '--terse',
            '--fields', 'DEVICE,TYPE',
            'device'
        ], stdout=PIPE)
        lines = proc.communicate()[0].decode('utf-8').splitlines()

        interfaces = []
        for line in lines:
            device, dtype = line.split(':')
            if dtype != 'wifi':
                continue
            interfaces.append(device)

        return interfaces

    def _is_enabled(self):
        '''
        Returns `True` if wifi is enabled else `False`.

        .. versionadded:: 1.2.5
        .. versionchanged:: 1.3.2
            nmcli output is properly decoded to unicode
        '''
        enbl = Popen(["nmcli", "radio", "wifi"], stdout=PIPE, stderr=PIPE)
        if enbl.communicate()[0].split()[0].decode('utf-8') == "enabled":
            return True
        return False

    def _is_connected(self, interface=None):
        '''
        .. versionadded:: 1.4.0
        '''

        if not interface:
            interface = self.interfaces[0]

        proc = Popen([
            'nmcli', '--terse',
            '--fields', 'DEVICE,TYPE,STATE',
            'device'
        ], stdout=PIPE)
        lines = proc.communicate()[0].decode('utf-8').splitlines()

        connected = False
        for line in lines:
            device, dtype, state = line.split(':')
            if dtype != 'wifi':
                continue

            if device != interface:
                continue

            if state == 'connected':
                connected = True

        return connected

    def _start_scanning(self, interface=None):
        '''
        Returns all the network information.

        .. versionadded:: 1.2.5
        .. versionchanged:: 1.3.0
            scan only if wifi is enabled
        '''

        if not interface:
            interface = self.interfaces[0]

        import wifi
        if self._is_enabled():
            list_ = list(wifi.Cell.all(interface))
            for i in range(len(list_)):
                self.names[list_[i].ssid] = list_[i]
        else:
            raise Exception('Wifi not enabled.')

    def _get_network_info(self, name):
        '''
        Starts scanning for available Wi-Fi networks and returns the available,
        devices.

        .. versionadded:: 1.2.5
        '''
        ret_list = {}
        ret_list['ssid'] = self.names[name].ssid
        ret_list['signal'] = self.names[name].signal
        ret_list['quality'] = self.names[name].quality
        ret_list['frequency'] = self.names[name].frequency
        ret_list['bitrates'] = self.names[name].bitrates
        ret_list['encrypted'] = self.names[name].encrypted
        ret_list['channel'] = self.names[name].channel
        ret_list['address'] = self.names[name].address
        ret_list['mode'] = self.names[name].mode
        if not ret_list['encrypted']:
            return ret_list
        else:
            ret_list['encryption_type'] = self.names[name].encryption_type
            return ret_list

    def _get_available_wifi(self):
        '''
        Returns the name of available networks.

        .. versionadded:: 1.2.5
        .. versionchanged:: 1.4.0
            return a proper list of elements instead of dict_keys
        '''
        return list(self.names.keys())

    def _connect(self, network, parameters, interface=None):
        '''
        Expects 2 parameters:
            - name/ssid of the network.
            - parameters: dict type
                - password: string or None

        .. versionadded:: 1.2.5
        '''
        if not interface:
            interface = self.interfaces[0]

        import wifi
        result = None
        try:
            self._enable()
        finally:
            password = parameters['password']
            cell = self.names[network]
            result = wifi.Scheme.for_cell(
                interface, network, cell, password
            )
        return result

    def _disconnect(self, interface=None):
        '''
        Disconnect all the networks managed by Network manager.

        .. versionadded:: 1.2.5
        '''
        if not interface:
            interface = self.interfaces[0]

        if self._nmcli_version() >= (1, 2, 6):
            call(['nmcli', 'dev', 'disconnect', interface])
        else:
            call(['nmcli', 'nm', 'enable', 'false'])

    def _enable(self):
        '''
        Wifi interface power state is set to "ON".

        .. versionadded:: 1.3.2
        '''
        return call(['nmcli', 'radio', 'wifi', 'on'])

    def _disable(self):
        '''
        Wifi interface power state is set to "OFF".

        .. versionadded:: 1.3.2
        '''
        return call(['nmcli', 'radio', 'wifi', 'off'])

    def _nmcli_version(self):
        '''
        .. versionadded:: 1.3.2
        '''
        version = Popen(['nmcli', '-v'], stdout=PIPE)
        version = version.communicate()[0].decode('utf-8')
        while version and not version[0].isdigit():
            version = version[1:]
        return tuple(map(int, (version.split('.'))))


def instance():
    import sys

    if whereis_exe('nmcli'):
        return NMCLIWifi()

    try:
        import wifi  # pylint: disable=unused-variable,unused-import
    except ImportError:
        sys.stderr.write("python-wifi not installed. try:"
                         "`pip install --user wifi`.")
        return Wifi()
    return LinuxWifi()



from jnius import autoclass
from jnius import cast
from jnius import java_method
from jnius import PythonJavaClass

from plyer.platforms.android import activity
from plyer.facades import Proximity

ActivityInfo = autoclass('android.content.pm.ActivityInfo')
Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class ProximitySensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(ProximitySensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_PROXIMITY)
        self.value = None

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.value = event.values[0]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidProximity(Proximity):

    listener = None

    def _enable(self, **kwargs):
        if not self.listener:
            self.listener = ProximitySensorListener()
            self.listener.enable()

    def _disable(self, **kwargs):
        if self.listener:
            self.listener.disable()
            delattr(self, 'listener')

    def _get_proximity(self):
        if self.listener:
            value = self.listener.value
            # value is 0.0 when proxime sensor is covered. In other case
            # value is 5.0 because in smartphone, optical proximity sensors
            # are used.
            return value < 5.0


def instance():
    return AndroidProximity()



''' Defines ctypes windows api.
'''

__all__ = ('GUID', 'get_DLLVERSIONINFO', 'MAKEDLLVERULL',
           'get_NOTIFYICONDATAW', 'CreateWindowExW', 'WindowProc',
           'DefWindowProcW', 'get_WNDCLASSEXW', 'GetModuleHandleW',
           'RegisterClassExW', 'UpdateWindow', 'LoadImageW',
           'Shell_NotifyIconW', 'DestroyIcon', 'UnregisterClassW',
           'DestroyWindow', 'LoadIconW', 'get_PATH')

import ctypes
from ctypes import Structure, windll, sizeof, POINTER, WINFUNCTYPE
from ctypes.wintypes import (
    DWORD, HICON, HWND, UINT, WCHAR, WORD, BYTE,
    LPCWSTR, INT, LPVOID, HINSTANCE, HMENU, LPARAM,
    WPARAM, HBRUSH, HMODULE, ATOM, BOOL, HANDLE
)
LRESULT = LPARAM
HRESULT = HANDLE
HCURSOR = HICON


class GUID(Structure):
    _fields_ = [
        ('Data1', DWORD),
        ('Data2', WORD),
        ('Data3', WORD),
        ('Data4', BYTE * 8)
    ]


class DLLVERSIONINFO(Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('dwMajorVersion', DWORD),
        ('dwMinorVersion', DWORD),
        ('dwBuildNumber', DWORD),
        ('dwPlatformID', DWORD),
    ]


def get_DLLVERSIONINFO(*largs):
    version_info = DLLVERSIONINFO(*largs)
    version_info.cbSize = sizeof(DLLVERSIONINFO)
    return version_info


def MAKEDLLVERULL(major, minor, build, sp):
    return (major << 48) | (minor << 32) | (build << 16) | sp


NOTIFYICONDATAW_fields = [
    ("cbSize", DWORD),
    ("hWnd", HWND),
    ("uID", UINT),
    ("uFlags", UINT),
    ("uCallbackMessage", UINT),
    ("hIcon", HICON),
    ("szTip", WCHAR * 128),
    ("dwState", DWORD),
    ("dwStateMask", DWORD),
    ("szInfo", WCHAR * 256),
    ("uVersion", UINT),
    ("szInfoTitle", WCHAR * 64),
    ("dwInfoFlags", DWORD),
    ("guidItem", GUID),
    ("hBalloonIcon", HICON),
]


class NOTIFYICONDATAW(Structure):
    _fields_ = NOTIFYICONDATAW_fields[:]


class NOTIFYICONDATAW_V3(Structure):
    _fields_ = NOTIFYICONDATAW_fields[:-1]


class NOTIFYICONDATAW_V2(Structure):
    _fields_ = NOTIFYICONDATAW_fields[:-2]


class NOTIFYICONDATAW_V1(Structure):
    _fields_ = NOTIFYICONDATAW_fields[:6]


NOTIFYICONDATA_V3_SIZE = sizeof(NOTIFYICONDATAW_V3)
NOTIFYICONDATA_V2_SIZE = sizeof(NOTIFYICONDATAW_V2)
NOTIFYICONDATA_V1_SIZE = sizeof(NOTIFYICONDATAW_V1)


def get_NOTIFYICONDATAW(*largs):
    notify_data = NOTIFYICONDATAW(*largs)

    # get shell32 version to find correct NOTIFYICONDATAW size
    DllGetVersion = windll.Shell32.DllGetVersion
    DllGetVersion.argtypes = [POINTER(DLLVERSIONINFO)]
    DllGetVersion.restype = HRESULT

    version = get_DLLVERSIONINFO()
    if DllGetVersion(version):
        raise Exception('Cannot get Windows version numbers.')
    v = MAKEDLLVERULL(version.dwMajorVersion, version.dwMinorVersion,
                      version.dwBuildNumber, version.dwPlatformID)

    # from the version info find the NOTIFYICONDATA size
    if v >= MAKEDLLVERULL(6, 0, 6, 0):
        notify_data.cbSize = sizeof(NOTIFYICONDATAW)
    elif v >= MAKEDLLVERULL(6, 0, 0, 0):
        notify_data.cbSize = NOTIFYICONDATA_V3_SIZE
    elif v >= MAKEDLLVERULL(5, 0, 0, 0):
        notify_data.cbSize = NOTIFYICONDATA_V2_SIZE
    else:
        notify_data.cbSize = NOTIFYICONDATA_V1_SIZE
    return notify_data


CreateWindowExW = windll.User32.CreateWindowExW
CreateWindowExW.argtypes = [DWORD, ATOM, LPCWSTR, DWORD, INT, INT, INT, INT,
                            HWND, HMENU, HINSTANCE, LPVOID]
CreateWindowExW.restype = HWND

GetModuleHandleW = windll.Kernel32.GetModuleHandleW
GetModuleHandleW.argtypes = [LPCWSTR]
GetModuleHandleW.restype = HMODULE

WindowProc = WINFUNCTYPE(LRESULT, HWND, UINT, WPARAM, LPARAM)
DefWindowProcW = windll.User32.DefWindowProcW
DefWindowProcW.argtypes = [HWND, UINT, WPARAM, LPARAM]
DefWindowProcW.restype = LRESULT


class WNDCLASSEXW(Structure):
    _fields_ = [
        ('cbSize', UINT),
        ('style', UINT),
        ('lpfnWndProc', WindowProc),
        ('cbClsExtra', INT),
        ('cbWndExtra', INT),
        ('hInstance', HINSTANCE),
        ('hIcon', HICON),
        ('hCursor', HCURSOR),
        ('hbrBackground', HBRUSH),
        ('lpszMenuName', LPCWSTR),
        ('lpszClassName', LPCWSTR),
        ('hIconSm', HICON),
    ]


def get_WNDCLASSEXW(*largs):
    wnd_class = WNDCLASSEXW(*largs)
    wnd_class.cbSize = sizeof(WNDCLASSEXW)
    return wnd_class


RegisterClassExW = windll.User32.RegisterClassExW
RegisterClassExW.argtypes = [POINTER(WNDCLASSEXW)]
RegisterClassExW.restype = ATOM

UpdateWindow = windll.User32.UpdateWindow
UpdateWindow.argtypes = [HWND]
UpdateWindow.restype = BOOL

LoadImageW = windll.User32.LoadImageW
LoadImageW.argtypes = [HINSTANCE, LPCWSTR, UINT, INT, INT, UINT]
LoadImageW.restype = HANDLE

Shell_NotifyIconW = windll.Shell32.Shell_NotifyIconW
Shell_NotifyIconW.argtypes = [DWORD, POINTER(NOTIFYICONDATAW)]
Shell_NotifyIconW.restype = BOOL

DestroyIcon = windll.User32.DestroyIcon
DestroyIcon.argtypes = [HICON]
DestroyIcon.restype = BOOL

UnregisterClassW = windll.User32.UnregisterClassW
UnregisterClassW.argtypes = [ATOM, HINSTANCE]
UnregisterClassW.restype = BOOL

DestroyWindow = windll.User32.DestroyWindow
DestroyWindow.argtypes = [HWND]
DestroyWindow.restype = BOOL

LoadIconW = windll.User32.LoadIconW
LoadIconW.argtypes = [HINSTANCE, LPCWSTR]
LoadIconW.restype = HICON


class SYSTEM_POWER_STATUS(Structure):
    _fields_ = [
        ('ACLineStatus', BYTE),
        ('BatteryFlag', BYTE),
        ('BatteryLifePercent', BYTE),
        ('Reserved1', BYTE),
        ('BatteryLifeTime', DWORD),
        ('BatteryFullLifeTime', DWORD),
    ]


SystemPowerStatusP = POINTER(SYSTEM_POWER_STATUS)

GetSystemPowerStatus = windll.kernel32.GetSystemPowerStatus
GetSystemPowerStatus.argtypes = [SystemPowerStatusP]
GetSystemPowerStatus.restype = BOOL


class GUID_(Structure):
    _fields_ = [
        ('Data1', DWORD),
        ('Data2', WORD),
        ('Data3', WORD),
        ('Data4', BYTE * 8)
    ]

    def __init__(self, uuid_):
        Structure.__init__(self)
        self.Data1, self.Data2, self.Data3, self.Data4[0], self.Data4[1], rest\
            = uuid_.fields
        for i in range(2, 8):
            self.Data4[i] = rest >> (8 - i - 1) * 8 & 0xff


_CoTaskMemFree = windll.ole32.CoTaskMemFree
_CoTaskMemFree.restype = None
_CoTaskMemFree.argtypes = [ctypes.c_void_p]

_SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath
_SHGetKnownFolderPath.argtypes = [
    POINTER(GUID_),
    DWORD,
    HANDLE,
    POINTER(ctypes.c_wchar_p)
]


class PathNotFoundException(Exception):
    pass


def get_PATH(folderid):
    fid = GUID_(folderid)
    pPath = ctypes.c_wchar_p()
    S_OK = 0
    Result = _SHGetKnownFolderPath(ctypes.byref(fid),
                                   0, None, ctypes.byref(pPath))
    if Result != S_OK:
        raise PathNotFoundException()
    path = pPath.value
    _CoTaskMemFree(pPath)
    return path



from plyer.facades import Keystore
from plyer.platforms.android import activity


class AndroidKeystore(Keystore):

    def _set_key(self, servicename, key, value, **kwargs):
        mode = kwargs.get("mode", 0)
        settings = activity.getSharedPreferences(servicename, mode)
        editor = settings.edit()
        editor.putString(key, value)
        editor.commit()

    def _get_key(self, servicename, key, **kwargs):
        mode = kwargs.get("mode", 0)
        default = kwargs.get("default", "__None")
        settings = activity.getSharedPreferences(servicename, mode)
        ret = settings.getString(key, default)
        if ret == "__None":
            ret = None
        return ret


def instance():
    return AndroidKeystore()



'''
Compatibility module for Python 2.7 and > 3.3
=============================================
'''
# pylint: disable=invalid-name

__all__ = ('PY2', 'string_types', 'queue', 'iterkeys',
           'itervalues', 'iteritems', 'xrange')

import sys
try:
    import queue
except ImportError:
    import Queue as queue

#: True if Python 2 intepreter is used
PY2 = sys.version_info[0] == 2

#: String types that can be used for checking if a object is a string
string_types = None
text_type = None
if PY2:
    # pylint: disable=undefined-variable
    # built-in actually, so it is defined in globals() for py2
    string_types = basestring
    text_type = unicode
else:
    string_types = text_type = str

if PY2:
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
else:
    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

if PY2:
    # pylint: disable=undefined-variable
    # built-in actually, so it is defined in globals() for py2
    xrange = xrange
else:
    xrange = range



from jnius import autoclass
from jnius import cast
from jnius import java_method
from jnius import PythonJavaClass
from math import exp
from plyer.facades import Humidity
from plyer.platforms.android import activity

ActivityInfo = autoclass('android.content.pm.ActivityInfo')
Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class RelativeHumiditySensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(RelativeHumiditySensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_RELATIVE_HUMIDITY)
        self.value = None

    def enable(self):
        self.SensorManager.registerListener(self, self.sensor,
                                            SensorManager.SENSOR_DELAY_NORMAL)

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.value = event.values[0]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AmbientTemperatureSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(AmbientTemperatureSensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_AMBIENT_TEMPERATURE)
        self.value = None

    def enable(self):
        self.SensorManager.registerListener(self, self.sensor,
                                            SensorManager.SENSOR_DELAY_NORMAL)

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.value = event.values[0]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidHumidity(Humidity):

    def __init__(self):
        self.state = False

    def _get_humidity(self):
        if self.state:
            m = 17.62
            Tn = 243.12
            Ta = 216.7
            Rh = self.listener_r.value
            Tc = self.listener_a.value
            A = 6.112
            K = 273.15
            humidity = (Ta * (Rh / 100) * A * exp(m * Tc / (Tn + Tc))
                        / (K + Tc))
            return humidity

    def _enable(self):
        if not self.state:
            self.listener_r = RelativeHumiditySensorListener()
            self.listener_a = AmbientTemperatureSensorListener()
            self.listener_r.enable()
            self.listener_a.enable()
            self.state = True

    def _disable(self):
        if self.state:
            self.listener_r.disable()
            self.listener_a.disable()
            self.state = False
            delattr(self, 'listener_r')
            delattr(self, 'listener_a')


def instance():
    return AndroidHumidity()



'''
Module of MacOS API for plyer.bluetooth.
'''

from subprocess import Popen, PIPE
from plyer.facades import Bluetooth
from plyer.utils import whereis_exe

from os import environ


class OSXBluetooth(Bluetooth):
    '''
    Implementation of MacOS bluetooth API.
    '''

    def _get_info(self):
        old_lang = environ.get('LANG')
        environ['LANG'] = 'C'

        sys_profiler_process = Popen(
            ["system_profiler", "SPBluetoothDataType"],
            stdout=PIPE
        )

        stdout = sys_profiler_process.communicate()[0].decode('utf-8')
        output = stdout.splitlines()

        lines = []
        for line in output:
            if 'Bluetooth Power' not in line:
                continue
            lines.append(line)

        if old_lang is None:
            environ.pop('LANG')
        else:
            environ['LANG'] = old_lang

        if output and len(lines) == 1:
            return lines[0].split()[2]
        else:
            return None


def instance():
    '''
    Instance for facade proxy.
    '''
    import sys
    if whereis_exe('system_profiler'):
        return OSXBluetooth()
    sys.stderr.write("system_profiler not found.")
    return Bluetooth()



'''
iOS GPS
-----------
'''

from pyobjus import autoclass, protocol
from pyobjus.dylib_manager import load_framework
from plyer.facades import GPS

load_framework('/System/Library/Frameworks/CoreLocation.framework')
CLLocationManager = autoclass('CLLocationManager')


class IosGPS(GPS):
    def _configure(self):
        if not hasattr(self, '_location_manager'):
            self._location_manager = CLLocationManager.alloc().init()

    def _start(self, **kwargs):
        self._location_manager.delegate = self

        self._location_manager.requestWhenInUseAuthorization()
        # NSLocationWhenInUseUsageDescription key must exist in Info.plist
        # file. When the authorization prompt is displayed your app goes
        # into pause mode and if your app doesn't support background mode
        # it will crash.
        self._location_manager.startUpdatingLocation()

    def _stop(self):
        self._location_manager.stopUpdatingLocation()

    @protocol('CLLocationManagerDelegate')
    def locationManager_didChangeAuthorizationStatus_(self, manager, status):
        if self.on_status:
            s_status = ''
            provider_status = ''
            provider = 'standard-ios-provider'
            if status == 0:
                provider_status = 'provider-disabled'
                s_status = 'notDetermined'
            elif status == 1:
                provider_status = 'provider-enabled'
                s_status = 'restricted'
            elif status == 2:
                provider_status = 'provider-disabled'
                s_status = 'denied'
            elif status == 3:
                provider_status = 'provider-enabled'
                s_status = 'authorizedAlways'
            elif status == 4:
                provider_status = 'provider-enabled'
                s_status = 'authorizedWhenInUse'
            self.on_status(provider_status, '{}: {}'.format(
                provider, s_status))

    @protocol('CLLocationManagerDelegate')
    def locationManager_didUpdateLocations_(self, manager, locations):
        location = manager.location

        description = location.description.UTF8String()
        split_description = description.split('<')[-1].split('>')[0].split(',')

        lat, lon = [float(coord) for coord in split_description]
        acc = float(description.split(' +/- ')[-1].split('m ')[0])

        speed = location.speed
        altitude = location.altitude
        course = location.course

        self.on_location(
            lat=lat,
            lon=lon,
            speed=speed,
            bearing=course,
            altitude=altitude,
            accuracy=acc)


def instance():
    return IosGPS()



class Humidity(object):
    '''Humidity facade.
       Humidity sensor returns value of humidity.
       With method `enable` you can turn on Humidity sensor and
       'disable' method stops the sensor.
       Use property `tell` to get humidity value.
    '''

    @property
    def tell(self):
        '''Current humidity'''
        return self._get_humidity()

    def enable(self):
        '''Enable Humidity sensor.'''
        self._enable()

    def disable(self):
        '''Disable Humidity sensor.'''
        self._disable()

    # private
    def _get_humidity(self, **kwargs):
        raise NotImplementedError()

    def _enable(self, **kwargs):
        raise NotImplementedError()

    def _disable(self, **kwargs):
        raise NotImplementedError()



from os import environ
from jnius import autoclass

ANDROID_VERSION = autoclass('android.os.Build$VERSION')
SDK_INT = ANDROID_VERSION.SDK_INT

try:
    from android import config
    ns = config.JAVA_NAMESPACE
except (ImportError, AttributeError):
    ns = 'org.renpy.android'

if 'PYTHON_SERVICE_ARGUMENT' in environ:
    PythonService = autoclass(ns + '.PythonService')
    activity = PythonService.mService
else:
    PythonActivity = autoclass(ns + '.PythonActivity')
    activity = PythonActivity.mActivity



'''
Speech to Text
==============

.. versionadded:: 1.4.0

Speech Recognition facade.

In order to check that your device supports voice recognition use method
`exist`.

Variable `language` indicates which language will be used to match words from
 voice.

Use `start` to start voice recognition immediately and `stop` to stop.

.. note::
    Needed permissions for Android: `RECORD_AUDIO` (and `INTERNET` if you want
    online voice recognition API to be used)

.. note::
    On Android platform, after execute `start` method you can hear BEEP!
    Mute sound in order to disable it.

.. note::
    For Android implementation to work there has to be an application with
    `android.speech.RecognitionService` implementation present in the system.
    Mostly it's `com.google.android.googlequicksearchbox` or "Google"
    application (the search bar with the launcher widget).

Offline Speech Recognition on Android
-------------------------------------

Requires any application that provides an
`android.speech.RecognitionService` implementation to the other apps. One of
such applications is on a lot of devices preinstalled Google (quick search
box).

The API prefers offline recognition, but should be able to switch to online
alternative in case you don't have a language package installed (`INTERNET`
permission necessary).

You can enable offline speech recognition this way (Android 8.1):

* open the `Settings` app
* choose `Language & Input` / `Language & Keyboard` (Samsung might include it
  in the `General` category)
* choose `On-Screen keyboard` or `Voice search`
* choose `Google Keyboard`
* choose `Offline Speech recognition`
* download language package if you don't have one already

Simple Examples
---------------

To start listening::

    >>> from plyer import stt
    >>> stt.start()

To retrieve partial results while listening::

    >>> assert stt.listening
    >>> print(stt.partial_results)

To stop listening::

    >>> stt.stop()

To retrieve results after the listening stopped::

    >>> print(stt.results)
'''


class STT(object):
    '''
    Speech to text facade.
    '''

    _language = 'en-US'
    '''
    Default language in which platform will try to recognize voice.
    In order to change language pick one from list by using
    `supported_languages` method.
    '''

    _supported_languages = [
        'en-US',
        'pl-PL'
    ]

    results = []
    '''
    List of sentences found while listening. It may consist of many similar
    and possible sentences that was recognition program.
    '''

    errors = []
    '''
    List of errors found while listening.
    '''

    partial_results = []
    '''
    List of results found while the listener is still being active.
    '''

    prefer_offline = True
    '''
    Preference whether to use offline language package necessary for
    each platform dependant implementation or online API.
    '''

    listening = False
    '''
    Current state of listening.
    '''

    @property
    def supported_languages(self):
        '''
        Return list of supported languages used in recognition.
        '''

        return self._supported_languages

    @property
    def language(self):
        '''
        Return current language.
        '''

        return self._language

    @language.setter
    def language(self, lang):
        '''
        Set current language.

        Value can not be set if it's not supported. See `supported_languages`
        to get what language you can set.

        .. note::
           We obviously can't check each language, therefore if you find
           that a specific language is available to you and the only limitation
           is our check for the internally defined `supported_languages`, feel
           free to open a pull request for adding your language to the list.
        '''

        if lang in self.supported_languages:
            self._language = lang

    # public methods
    def start(self):
        '''
        Start listening.
        '''

        self.results = []
        self.partial_results = []
        self._start()
        self.listening = True

    def stop(self):
        '''
        Stop listening.
        '''

        self._stop()
        self.listening = False

    def exist(self):
        '''
        Returns a boolean for speech recognition availability.
        '''

        return self._exist()

    # private methods
    def _start(self):
        raise NotImplementedError

    def _stop(self):
        raise NotImplementedError

    def _exist(self):
        raise NotImplementedError



'''
Android Gyroscope
-----------------
'''

from plyer.facades import Gyroscope
from jnius import PythonJavaClass, java_method, autoclass, cast
from plyer.platforms.android import activity

Context = autoclass('android.content.Context')
Sensor = autoclass('android.hardware.Sensor')
SensorManager = autoclass('android.hardware.SensorManager')


class GyroscopeSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(GyroscopeSensorListener, self).__init__()
        self.SensorManager = cast(
            'android.hardware.SensorManager',
            activity.getSystemService(Context.SENSOR_SERVICE)
        )
        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_GYROSCOPE
        )

        self.values = [None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:3]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        # Maybe, do something in future?
        pass


class GyroUncalibratedSensorListener(PythonJavaClass):
    __javainterfaces__ = ['android/hardware/SensorEventListener']

    def __init__(self):
        super(GyroUncalibratedSensorListener, self).__init__()
        service = activity.getSystemService(Context.SENSOR_SERVICE)
        self.SensorManager = cast('android.hardware.SensorManager', service)

        self.sensor = self.SensorManager.getDefaultSensor(
            Sensor.TYPE_GYROSCOPE_UNCALIBRATED)
        self.values = [None, None, None, None, None, None]

    def enable(self):
        self.SensorManager.registerListener(
            self, self.sensor,
            SensorManager.SENSOR_DELAY_NORMAL
        )

    def disable(self):
        self.SensorManager.unregisterListener(self, self.sensor)

    @java_method('(Landroid/hardware/SensorEvent;)V')
    def onSensorChanged(self, event):
        self.values = event.values[:6]

    @java_method('(Landroid/hardware/Sensor;I)V')
    def onAccuracyChanged(self, sensor, accuracy):
        pass


class AndroidGyroscope(Gyroscope):
    def __init__(self):
        super(AndroidGyroscope, self).__init__()
        self.bState = False

    def _enable(self):
        if (not self.bState):
            self.listenerg = GyroscopeSensorListener()
            self.listenergu = GyroUncalibratedSensorListener()
            self.listenerg.enable()
            self.listenergu.enable()
            self.bState = True

    def _disable(self):
        if (self.bState):
            self.bState = False
            self.listenerg.disable()
            self.listenergu.disable()
            del self.listenerg
            del self.listenergu

    def _get_orientation(self):
        if (self.bState):
            return tuple(self.listenerg.values)
        else:
            return (None, None, None)

    def _get_rotation_uncalib(self):
        if (self.bState):
            return tuple(self.listenergu.values)
        else:
            return (None, None, None, None, None, None)

    def __del__(self):
        if(self.bState):
            self._disable()
        super(self.__class__, self).__del__()


def instance():
    return AndroidGyroscope()



'''
Windows Storage Path
--------------------
'''

from plyer.facades import StoragePath
from os.path import expanduser
from plyer.platforms.win.libs.win_api_defs import get_PATH
from uuid import UUID


class WinStoragePath(StoragePath):

    def _get_home_dir(self):
        return expanduser('~')

    def _get_external_storage_dir(self):
        '''
        To do.
        '''
        return "Method not implemented for current platform."

    def _get_root_dir(self):
        folderid = UUID('{F38BF404-1D43-42F2-9305-67DE0B28FC23}')
        return get_PATH(folderid)

    def _get_documents_dir(self):
        folderid = UUID('{FDD39AD0-238F-46AF-ADB4-6C85480369C7}')
        return get_PATH(folderid)

    def _get_downloads_dir(self):
        folderid = UUID('{374DE290-123F-4565-9164-39C4925E467B}')
        return get_PATH(folderid)

    def _get_videos_dir(self):
        folderid = UUID('{18989B1D-99B5-455B-841C-AB7C74E4DDFC}')
        return get_PATH(folderid)

    def _get_music_dir(self):
        folderid = UUID('{4BD8D571-6D19-48D3-BE97-422220080E43}')
        return get_PATH(folderid)

    def _get_pictures_dir(self):
        folderid = UUID('{33E28130-4E1E-4676-835A-98395C3BC3BB}')
        return get_PATH(folderid)

    def _get_application_dir(self):
        folderid = UUID('{905E63B6-C1BF-494E-B29C-65B732D3D21A}')
        return get_PATH(folderid)


def instance():
    return WinStoragePath()



from pyobjus import autoclass
from pyobjus.dylib_manager import load_framework, INCLUDE

from plyer.facades import Wifi

load_framework(INCLUDE.Foundation)
load_framework(INCLUDE.CoreWLAN)

CWInterface = autoclass('CWInterface')
CWNetwork = autoclass('CWNetwork')
CWWiFiClient = autoclass('CWWiFiClient')
NSArray = autoclass('NSArray')
NSDictionary = autoclass('NSDictionary')
NSString = autoclass('NSString')


class OSXWifi(Wifi):
    names = {}

    def _is_enabled(self):
        '''
        Returns `True` if the Wifi is enabled else  returns `False`.
        '''
        return CWWiFiClient.sharedWiFiClient().interface().powerOn()

    def _get_network_info(self, name):
        '''
        Returns all the network information.
        '''

        accessNetworkType = self.names[name].accessNetworkType
        aggregateRSSI = self.names[name].aggregateRSSI
        beaconInterval = self.names[name].beaconInterval
        bssid = self.names[name].bssid.UTF8String()
        countryCode = self.names[name].countryCode
        hasInternet = self.names[name].hasInternet
        hasInterworkingIE = self.names[name].hasInterworkingIE
        hessid = self.names[name].hessid
        ibss = self.names[name].ibss
        isAdditionalStepRequiredForAccess = \
            self.names[name].isAdditionalStepRequiredForAccess
        isCarPlayNetwork = self.names[name].isCarPlayNetwork
        isEmergencyServicesReachable = \
            self.names[name].isEmergencyServicesReachable
        isPasspoint = self.names[name].isPasspoint
        isPersonalHotspot = self.names[name].isPersonalHotspot
        isUnauthenticatedEmergencyServiceAccessible = \
            self.names[name].isUnauthenticatedEmergencyServiceAccessible
        noiseMeasurement = self.names[name].noiseMeasurement
        physicalLayerMode = self.names[name].physicalLayerMode
        rssiValue = self.names[name].rssiValue
        securityType = self.names[name].securityType
        ssid = self.names[name].ssid.UTF8String()
        supportsEasyConnect = self.names[name].supportsEasyConnect
        supportsWPS = self.names[name].supportsWPS
        venueGroup = self.names[name].venueGroup
        venueType = self.names[name].venueType

        return {'accessNetworkType': accessNetworkType,
                'aggregateRSSI': aggregateRSSI,
                'beaconInterval': beaconInterval,
                'bssid': bssid,
                'countryCode': countryCode,
                'hasInternet': hasInternet,
                'hasInterworkingIE': hasInterworkingIE,
                'hessid': hessid,
                'ibss': ibss,
                'isAdditionalStepRequiredForAccess':
                isAdditionalStepRequiredForAccess,
                'isCarPlayNetwork': isCarPlayNetwork,
                'isEmergencyServicesReachable': isEmergencyServicesReachable,
                'isPasspoint': isPasspoint,
                'isPersonalHotspot': isPersonalHotspot,
                'isUnauthenticatedEmergencyServiceAccessible':
                isUnauthenticatedEmergencyServiceAccessible,
                'noiseMeasurement': noiseMeasurement,
                'physicalLayerMode': physicalLayerMode,
                'rssiValue': rssiValue,
                'securityType': securityType,
                'ssid': ssid,
                'supportsEasyConnect': supportsEasyConnect,
                'supportsWPS': supportsWPS,
                'venueGroup': venueGroup,
                'venueType': venueType}

    def _start_scanning(self):
        '''
        Starts scanning for available Wi-Fi networks.
        '''
        if self._is_enabled():
            self.names = {}
            c = CWInterface.interface()
            scan = c.scanForNetworksWithName_error_(None, None)
            cnt = scan.allObjects().count()
            for i in range(cnt):
                self.names[
                    scan.allObjects().objectAtIndex_(i).ssid.UTF8String()
                ] = scan.allObjects().objectAtIndex_(i)
        else:
            raise Exception("Wifi not enabled.")

    def _get_available_wifi(self):
        '''
        Returns the name of available networks.
        '''
        return self.names.keys()

    def _connect(self, network, parameters):
        '''
        Expects 2 parameters:
            - name/ssid of the network.
            - password: dict type
        '''
        password = parameters['password']
        network_object = self.names[network]
        CWInterface.interface().associateToNetwork_password_error_(
            network_object,
            password,
            None)
        return

    def _disconnect(self):
        '''
        Disconnect from network.
        '''
        CWInterface.interface().disassociate()
        return

    def _disable(self):
        '''
        Wifi interface power state is set to "OFF".
        '''

        interface = CWWiFiClient.sharedWiFiClient().interface()
        interface.setPower_error_(False, None)

    def _enable(self):
        '''
        Wifi interface power state is set to "ON".
        '''

        interface = CWWiFiClient.sharedWiFiClient().interface()
        interface.setPower_error_(True, None)


def instance():
    return OSXWifi()



'''
Linux Storage Path
--------------------
'''

from plyer.facades import StoragePath
from os.path import expanduser, dirname, abspath

# Default paths for each name
USER_DIRS = "/.config/user-dirs.dirs"

PATHS = {
    "DESKTOP": "/Desktop",
    "DOCUMENTS": "/Documents",
    "DOWNLOAD": "/Downloads",
    "MUSIC": "/Music",
    "PICTURES": "/Pictures",
    "VIDEOS": "/Videos"
}


class LinuxStoragePath(StoragePath):

    def _get_from_user_dirs(self, name):
        try:
            with open(self._get_home_dir() + USER_DIRS, "r") as f:
                lines = f.readlines()
                # Find the line that starts with XDG_<name> to get the path
                index = [i for i, v in enumerate(lines)
                         if v.startswith("XDG_" + name)][0]
                return lines[index].split('"')[1]
        except KeyError:
            return PATHS[name]
        except Exception as e:
            raise e

    def _get_home_dir(self):
        return expanduser('~')

    def _get_external_storage_dir(self):
        return "/media/" + self._get_home_dir().split("/")[-1]

    def _get_root_dir(self):
        return "/"

    def _get_documents_dir(self):
        directory = self._get_from_user_dirs("DOCUMENT")
        return directory.replace("$HOME", self._get_home_dir())

    def _get_downloads_dir(self):
        directory = self._get_from_user_dirs("DOWNLOAD")
        return directory.replace("$HOME", self._get_home_dir())

    def _get_videos_dir(self):
        directory = self._get_from_user_dirs("VIDEOS")
        return directory.replace("$HOME", self._get_home_dir())

    def _get_music_dir(self):
        directory = self._get_from_user_dirs("MUSIC")
        return directory.replace("$HOME", self._get_home_dir())

    def _get_pictures_dir(self):
        directory = self._get_from_user_dirs("PICTURES")
        return directory.replace("$HOME", self._get_home_dir())

    def _get_application_dir(self):
        return dirname(abspath(__name__))


def instance():
    return LinuxStoragePath()



from android.runnable import run_on_ui_thread

from jnius import autoclass
from jnius import java_method
from jnius import PythonJavaClass

from plyer.facades import STT
from plyer.platforms.android import activity

ArrayList = autoclass('java.util.ArrayList')
Bundle = autoclass('android.os.Bundle')
Context = autoclass('android.content.Context')
Intent = autoclass('android.content.Intent')
RecognizerIntent = autoclass('android.speech.RecognizerIntent')
RecognitionListener = autoclass('android.speech.RecognitionListener')
SpeechRecognizer = autoclass('android.speech.SpeechRecognizer')

SpeechResults = SpeechRecognizer.RESULTS_RECOGNITION


class SpeechListener(PythonJavaClass):
    __javainterfaces__ = ['android/speech/RecognitionListener']

    # class variables because PythonJavaClass class failed
    # to see them later in getters and setters
    _error_callback = None
    _result_callback = None
    _partial_result_callback = None
    _volume_callback = None

    def __init__(self):
        super(SpeechListener, self).__init__()

        # overwrite class variables in the object
        self._error_callback = None
        self._result_callback = None
        self._partial_result_callback = None
        self._volume_callback = None

    # error handling
    @property
    def error_callback(self):
        return self._error_callback

    @error_callback.setter
    def error_callback(self, callback):
        '''
        Set error callback. It is called when error occurs.

        :param callback: function with one parameter for error message
        '''

        self._error_callback = callback

    # result handling
    @property
    def result_callback(self):
        return self._result_callback

    @result_callback.setter
    def result_callback(self, callback):
        '''
        Set result callback. It is called when results are received.

        :param callback: function with one parameter for lists of strings
        '''

        self._result_callback = callback

    @property
    def partial_result_callback(self):
        return self._partial_result_callback

    @partial_result_callback.setter
    def partial_result_callback(self, callback):
        '''
        Set partial result callback. It is called when partial results are
        received while the listener is still in listening mode.

        :param callback: function with one parameter for lists of strings
        '''

        self._partial_result_callback = callback

    # voice changes handling
    @property
    def volume_callback(self):
        return self._volume_callback

    @volume_callback.setter
    def volume_callback(self, callback):
        '''
        Set volume voice callback.

        It is called when loudness of the voice changes.

        :param callback: function with one parameter for volume RMS dB (float).
        '''
        self._volume_callback = callback

    # Implementation Java Interfaces
    @java_method('()V')
    def onBeginningOfSpeech(self):
        pass

    @java_method('([B)V')
    def onBufferReceived(self, buffer):
        pass

    @java_method('()V')
    def onEndOfSpeech(self):
        pass

    @java_method('(I)V')
    def onError(self, error):
        msg = ''
        if error == SpeechRecognizer.ERROR_AUDIO:
            msg = 'audio'
        if error == SpeechRecognizer.ERROR_CLIENT:
            msg = 'client'
        if error == SpeechRecognizer.ERROR_INSUFFICIENT_PERMISSIONS:
            msg = 'insufficient_permissions'
        if error == SpeechRecognizer.ERROR_NETWORK:
            msg = 'network'
        if error == SpeechRecognizer.ERROR_NETWORK_TIMEOUT:
            msg = 'network_timeout'
        if error == SpeechRecognizer.ERROR_NO_MATCH:
            msg = 'no_match'
        if error == SpeechRecognizer.ERROR_RECOGNIZER_BUSY:
            msg = 'recognizer_busy'
        if error == SpeechRecognizer.ERROR_SERVER:
            msg = 'server'
        if error == SpeechRecognizer.ERROR_SPEECH_TIMEOUT:
            msg = 'speech_timeout'

        if msg and self.error_callback:
            self.error_callback('error:' + msg)

    @java_method('(ILandroid/os/Bundle;)V')
    def onEvent(self, event_type, params):
        pass

    @java_method('(Landroid/os/Bundle;)V')
    def onPartialResults(self, results):
        texts = []
        matches = results.getStringArrayList(SpeechResults)
        for match in matches.toArray():
            if isinstance(match, bytes):
                match = match.decode('utf-8')
            texts.append(match)

        if texts and self.partial_result_callback:
            self.partial_result_callback(texts)

    @java_method('(Landroid/os/Bundle;)V')
    def onReadyForSpeech(self, params):
        pass

    @java_method('(Landroid/os/Bundle;)V')
    def onResults(self, results):
        texts = []
        matches = results.getStringArrayList(SpeechResults)
        for match in matches.toArray():
            if isinstance(match, bytes):
                match = match.decode('utf-8')
            texts.append(match)

        if texts and self.result_callback:
            self.result_callback(texts)

    @java_method('(F)V')
    def onRmsChanged(self, rmsdB):
        if self.volume_callback:
            self.volume_callback(rmsdB)


class AndroidSpeech(STT):
    '''
    Android Speech Implementation.

    Android class `SpeechRecognizer`'s listening deactivates automatically.

    Class methods `_on_error()`, `_on_result()` listeners. You can find
    documentation here:
    https://developer.android.com/reference/android/speech/RecognitionListener
    '''

    def _on_error(self, msg):
        self.errors.append(msg)
        self.stop()

    def _on_result(self, messages):
        self.results.extend(messages)
        self.stop()

    def _on_partial(self, messages):
        self.partial_results.extend(messages)

    @run_on_ui_thread
    def _start(self):
        intent = Intent(RecognizerIntent.ACTION_RECOGNIZE_SPEECH)
        intent.putExtra(
            RecognizerIntent.EXTRA_CALLING_PACKAGE,
            activity.getPackageName()
        )

        # language preferences
        intent.putExtra(
            RecognizerIntent.EXTRA_LANGUAGE_PREFERENCE, self.language
        )
        intent.putExtra(
            RecognizerIntent.EXTRA_LANGUAGE_MODEL,
            RecognizerIntent.LANGUAGE_MODEL_WEB_SEARCH
        )

        # results settings
        intent.putExtra(RecognizerIntent.EXTRA_MAX_RESULTS, 1000)
        intent.putExtra(RecognizerIntent.EXTRA_PARTIAL_RESULTS, True)
        if self.prefer_offline:
            intent.putExtra(RecognizerIntent.EXTRA_PREFER_OFFLINE, True)

        # listener and callbacks
        listener = SpeechListener()
        listener.error_callback = self._on_error
        listener.result_callback = self._on_result
        listener.partial_result_callback = self._on_partial

        # create recognizer and start
        self.speech = SpeechRecognizer.createSpeechRecognizer(activity)
        self.speech.setRecognitionListener(listener)
        self.speech.startListening(intent)

    @run_on_ui_thread
    def _stop(self):
        if not self.speech:
            return

        # stop listening
        self.speech.stopListening()

        # free object
        self.speech.destroy()
        self.speech = None

    def _exist(self):
        return bool(
            SpeechRecognizer.isRecognitionAvailable(activity)
        )


def instance():
    return AndroidSpeech()



'''
Documentation:
http://docs.microsoft.com/en-us/windows/desktop/Multimedia

.. versionadded:: 1.4.0
'''

from __future__ import unicode_literals
from os.path import join

from ctypes import windll
from ctypes import (
    sizeof, c_void_p, c_ulonglong, c_ulong,
    c_wchar_p, byref, Structure, create_string_buffer
)
from ctypes.wintypes import DWORD, UINT

from plyer.facades import Audio
from plyer.platforms.win.storagepath import WinStoragePath

# DWORD_PTR i.e. ULONG_PTR, 32/64bit
ULONG_PTR = c_ulonglong if sizeof(c_void_p) == 8 else c_ulong

# device specific symbols
MCI_OPEN = 0x803
MCI_OPEN_TYPE = 0x2000
MCI_OPEN_ELEMENT = 512
MCI_RECORD = 0x80F
MCI_STOP = 0x808
MCI_SAVE = 0x813
MCI_PLAY = 0x806
MCI_CLOSE = 0x804

# recorder specific symbols
MCI_FROM = 4
MCI_TO = 8
MCI_WAIT = 2
MCI_SAVE_FILE = 256


class MCI_OPEN_PARMS(Structure):
    '''
    Struct for MCI_OPEN message parameters.

    .. versionadded:: 1.4.0
    '''

    _fields_ = [
        ('mciOpenParms', ULONG_PTR),
        ('wDeviceID', UINT),
        ('lpstrDeviceType', c_wchar_p),
        ('lpstrElementName', c_wchar_p),
        ('lpstrAlias', c_wchar_p)
    ]


class MCI_RECORD_PARMS(Structure):
    '''
    Struct for MCI_RECORD message parameters.

    http://docs.microsoft.com/en-us/windows/desktop/Multimedia/mci-record-parms

    .. versionadded:: 1.4.0
    '''

    _fields_ = [
        ('dwCallback', ULONG_PTR),
        ('dwFrom', DWORD),
        ('dwTo', DWORD)
    ]


class MCI_SAVE_PARMS(Structure):
    '''
    Struct for MCI_SAVE message parameters.

    http://docs.microsoft.com/en-us/windows/desktop/Multimedia/mci-save-parms

    .. versionadded:: 1.4.0
    '''

    _fields_ = [
        ('dwCallback', ULONG_PTR),
        ('lpfilename', c_wchar_p)
    ]


class MCI_PLAY_PARMS(Structure):
    '''
    Struct for MCI_PLAY message parameters.

    http://docs.microsoft.com/en-us/windows/desktop/Multimedia/mci-play-parms

    .. versionadded:: 1.4.0
    '''

    _fields_ = [
        ('dwCallback', ULONG_PTR),
        ('dwFrom', DWORD),
        ('dwTo', DWORD)
    ]


def send_command(device, msg, flags, params):
    '''
    Generic mciSendCommandW() wrapper with error handler.
    All parameters are required as for mciSendCommandW().
    In case of no `params` passed, use `None`, that value
    won't be dereferenced.

    .. versionadded:: 1.4.0
    '''

    multimedia = windll.winmm
    send_command_w = multimedia.mciSendCommandW
    get_error = multimedia.mciGetErrorStringW

    # error text buffer
    # by API specification 128 is max, however the API sometimes
    # kind of does not respect the documented bounds and returns
    # more characters than buffer length...?!
    error_len = 128

    # big enough to prevent API accidentally segfaulting
    error_text = create_string_buffer(error_len * 2)

    # open a recording device with a new file
    error_code = send_command_w(
        device,  # device ID
        msg,
        flags,

        # reference to parameters structure or original value
        # in case of params=False/0/None/...
        byref(params) if params else params
    )

    # handle error messages if any
    if error_code:
        # device did not open, raise an exception
        get_error(error_code, byref(error_text), error_len)
        error_text = error_text.raw.replace(b'\x00', b'').decode('utf-8')

        # either it can close already open device or it will fail because
        # the device is in non-closable state, but the end result is the same
        # and it makes no sense to parse MCI_CLOSE's error in this case
        send_command_w(device, MCI_CLOSE, 0, None)
        raise Exception(error_code, error_text)

    # return params struct because some commands write into it
    # to pass some values out of the local function scope
    return params


class WinRecorder(object):
    '''
    Generic wrapper for MCI_RECORD handling the filenames and device closing
    in the same approach like it is used for other platforms.

    .. versionadded:: 1.4.0
    '''

    def __init__(self, device, filename):
        self._device = device
        self._filename = filename

    @property
    def device(self):
        '''
        Public property returning device ID.

        .. versionadded:: 1.4.0
        '''
        return self._device

    @property
    def filename(self):
        '''
        Public property returning filename for current recording.

        .. versionadded:: 1.4.0
        '''
        return self._filename

    def record(self):
        '''
        Start recording a WAV sound.

        .. versionadded:: 1.4.0
        '''
        send_command(
            device=self.device,
            msg=MCI_RECORD,
            flags=0,
            params=None
        )

    def stop(self):
        '''
        Stop recording and save the data to a file path
        self.filename. Wait until the file is written.
        Close the device afterwards.

        .. versionadded:: 1.4.0
        '''

        # stop the recording first
        send_command(
            device=self.device,
            msg=MCI_STOP,
            flags=MCI_WAIT,
            params=None
        )

        # choose filename for the WAV file
        save_params = MCI_SAVE_PARMS()
        save_params.lpfilename = self.filename

        # save the sound data to a file and wait
        # until it ends writing to the file
        send_command(
            device=self.device,
            msg=MCI_SAVE,
            flags=MCI_SAVE_FILE | MCI_WAIT,
            params=save_params
        )

        # close the recording device
        send_command(
            device=self.device,
            msg=MCI_CLOSE,
            flags=0,
            params=None
        )


class WinPlayer(object):
    '''
    Generic wrapper for MCI_PLAY handling the device closing.

    .. versionadded:: 1.4.0
    '''

    def __init__(self, device):
        self._device = device

    @property
    def device(self):
        '''
        Public property returning device ID.

        .. versionadded:: 1.4.0
        '''
        return self._device

    def play(self):
        '''
        Start playing a WAV sound.

        .. versionadded:: 1.4.0
        '''
        play_params = MCI_PLAY_PARMS()
        play_params.dwFrom = 0

        send_command(
            device=self.device,
            msg=MCI_PLAY,
            flags=MCI_FROM,
            params=play_params
        )

    def stop(self):
        '''
        Stop playing a WAV sound and close the device.

        .. versionadded:: 1.4.0
        '''
        send_command(
            device=self.device,
            msg=MCI_STOP,
            flags=MCI_WAIT,
            params=None
        )

        # close the playing device
        send_command(
            device=self.device,
            msg=MCI_CLOSE,
            flags=0,
            params=None
        )


class WinAudio(Audio):
    '''
    Windows implementation of audio recording and audio playing.

    .. versionadded:: 1.4.0
    '''

    def __init__(self, file_path=None):
        # default path unless specified otherwise
        default_path = join(
            WinStoragePath().get_music_dir(),
            'audio.wav'
        )
        super(WinAudio, self).__init__(file_path or default_path)

        self._recorder = None
        self._player = None
        self._current_file = None

    def _start(self):
        '''
        Start recording a WAV sound in the background asynchronously.

        .. versionadded:: 1.4.0
        '''

        # clean everything before recording in case
        # there is a different device open
        self._stop()

        # create structure and set device parameters
        open_params = MCI_OPEN_PARMS()
        open_params.lpstrDeviceType = 'waveaudio'
        open_params.lpstrElementName = ''

        # open a new device for recording
        open_params = send_command(
            device=0,  # device ID before opening
            msg=MCI_OPEN,

            # empty filename in lpstrElementName
            # device type in lpstrDeviceType
            flags=MCI_OPEN_ELEMENT | MCI_OPEN_TYPE,
            params=open_params
        )

        # get recorder with device id and path for saving
        self._recorder = WinRecorder(
            device=open_params.wDeviceID,
            filename=self._file_path
        )
        self._recorder.record()

        # Setting the currently recorded file as current file
        # for using it as a parameter in audio player
        self._current_file = self._recorder.filename

    def _stop(self):
        '''
        Stop recording or playing of a WAV sound.

        .. versionadded:: 1.4.0
        '''

        if self._recorder:
            self._recorder.stop()
            self._recorder = None

        if self._player:
            self._player.stop()
            self._player = None

    def _play(self):
        '''
        Play a WAV sound from a file. Prioritize latest recorded file before
        default file path from WinAudio.

        .. versionadded:: 1.4.0
        '''

        # create structure and set device parameters
        open_params = MCI_OPEN_PARMS()
        open_params.lpstrDeviceType = 'waveaudio'
        open_params.lpstrElementName = self._current_file or self._file_path

        # open a new device for playing
        open_params = send_command(
            device=0,  # device ID before opening
            msg=MCI_OPEN,

            # existing filename in lpstrElementName
            # device type in lpstrDeviceType
            flags=MCI_OPEN_ELEMENT | MCI_OPEN_TYPE,
            params=open_params
        )

        # get recorder with device id and path for saving
        self._player = WinPlayer(device=open_params.wDeviceID)
        self._player.play()


def instance():
    '''
    Instance for facade proxy.
    '''
    return WinAudio()



'''
iOS Brightness
--------------
'''

from pyobjus import autoclass
from plyer.facades import Brightness
from pyobjus.dylib_manager import load_framework

load_framework('/System/Library/Frameworks/UIKit.framework')
UIScreen = autoclass('UIScreen')


class iOSBrightness(Brightness):

    def __init__(self):
        self.screen = UIScreen.mainScreen()

    def _current_level(self):
        return self.screen.brightness * 100

    def set_level(self, level):
        self.screen.brightness = level / 100


def instance():
    return iOSBrightness()



'''
Module of Windows API for plyer.email.
'''

import os
try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote
from plyer.facades import Email


class WindowsEmail(Email):
    # pylint: disable=too-few-public-methods
    '''
    Implementation of Windows email API.
    '''

    def _send(self, **kwargs):
        recipient = kwargs.get('recipient')
        subject = kwargs.get('subject')
        text = kwargs.get('text')

        uri = "mailto:"
        if recipient:
            uri += str(recipient)
        if subject:
            uri += "?" if "?" not in uri else "&"
            uri += "subject="
            uri += quote(str(subject))
        if text:
            uri += "?" if "?" not in uri else "&"
            uri += "body="
            uri += quote(str(text))

        # WE + startfile are available only on Windows
        try:
            os.startfile(uri)  # pylint: disable=no-member
        except WindowsError:  # pylint: disable=undefined-variable
            print("Warning: unable to find a program able to send emails.")


def instance():
    '''
    Instance for facade proxy.
    '''
    return WindowsEmail()
